!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_FAST_RG
INTERFACE
SUBROUTINE ICE4_FAST_RG(LDSOFT, PCOMPUTE, KRR, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, PCIT, &
                       &PLBDAR, PLBDAS, PLBDAG, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, &
                       &PRGSI, PRGSI_MR, &
                       &PWETG, &
                       &PRICFRRG, PRRCFRIG, PRICFRR, PRCWETG, PRIWETG, PRRWETG, PRSWETG, &
                       &PRCDRYG, PRIDRYG, PRRDRYG, PRSDRYG, PRWETGH, PRWETGH_MR, PRGMLTR, &
                       &PRG_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH, PB_RG, PB_RH)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
INTEGER,                  INTENT(IN)    :: KRR      ! Number of moist variable
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGSI    ! Graupel tendency by other processes
REAL, DIMENSION(:),       INTENT(IN)    :: PRGSI_MR ! Graupel mr change by other processes
REAL, DIMENSION(:),       INTENT(OUT)   :: PWETG    ! True where graupel grows in wet mode
REAL, DIMENSION(:),       INTENT(INOUT) :: PRICFRRG ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PRRCFRIG ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PRICFRR  ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRWETGH  ! Conversion of graupel into hail
REAL, DIMENSION(:),       INTENT(OUT)   :: PRWETGH_MR ! Conversion of graupel into hail, mr change
REAL, DIMENSION(:),       INTENT(INOUT) :: PRGMLTR  ! Melting of the graupel
REAL, DIMENSION(:, :),    INTENT(INOUT) :: PRG_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RH
END SUBROUTINE ICE4_FAST_RG
END INTERFACE
END MODULE MODI_ICE4_FAST_RG
SUBROUTINE ICE4_FAST_RG(LDSOFT, PCOMPUTE, KRR, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, PCIT, &
                       &PLBDAR, PLBDAS, PLBDAG, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, &
                       &PRGSI, PRGSI_MR, &
                       &PWETG, &
                       &PRICFRRG, PRRCFRIG, PRICFRR, PRCWETG, PRIWETG, PRRWETG, PRSWETG, &
                       &PRCDRYG, PRIDRYG, PRRDRYG, PRSDRYG, PRWETGH, PRWETGH_MR, PRGMLTR, &
                       &PRG_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH, PB_RG, PB_RH)
!!
!!**  PURPOSE
!!    -------
!!      Computes the fast rg processes
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XGAMW,XCI,XCL,XCPV,XESTT,XGAMI,XLMTT,XLVTT,XMD,XMV,XRV,XTT,   &
                               XEPSILO
USE MODD_PARAM_ICE,      ONLY: LCRFLIMIT,LEVLIMIT,LNULLWETG,LWETGPOST
USE MODD_RAIN_ICE_DESCR, ONLY: XBS,XCEXVT,XCXG,XCXS,XDG,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: NDRYLBDAG,NDRYLBDAR,NDRYLBDAS,X0DEPG,X1DEPG,XCOLEXIG,XCOLEXSG,XCOLIG,XCOLSG,XDRYINTP1G, &
                               XDRYINTP1R,XDRYINTP1S,XDRYINTP2G,XDRYINTP2R,XDRYINTP2S,XEX0DEPG,XEX1DEPG,XEXICFRR,      &
                               XEXRCFRI,XFCDRYG,XFIDRYG,XFRDRYG,XFSDRYG,XICFRR,XKER_RDRYG,XKER_SDRYG,XLBRDRYG1,        &
                               XLBRDRYG2,XLBRDRYG3,XLBSDRYG1,XLBSDRYG2,XLBSDRYG3,XRCFRI

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB

#ifdef MNH_BITREP
USE MODI_BITREP
#endif

IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
INTEGER,                  INTENT(IN)    :: KRR      ! Number of moist variable
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGSI    ! Graupel tendency by other processes
REAL, DIMENSION(:),       INTENT(IN)    :: PRGSI_MR ! Graupel mr change by other processes
REAL, DIMENSION(:),       INTENT(OUT)   :: PWETG    ! True where graupel grows in wet mode
REAL, DIMENSION(:),       INTENT(INOUT) :: PRICFRRG ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PRRCFRIG ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(INOUT) :: PRICFRR  ! Rain contact freezing
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSWETG  ! Graupel wet growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSDRYG  ! Graupel dry growth
REAL, DIMENSION(:),       INTENT(OUT)   :: PRWETGH  ! Conversion of graupel into hail
REAL, DIMENSION(:),       INTENT(OUT)   :: PRWETGH_MR ! Conversion of graupel into hail, mr change
REAL, DIMENSION(:),       INTENT(INOUT) :: PRGMLTR  ! Melting of the graupel
REAL, DIMENSION(:, :),    INTENT(INOUT) :: PRG_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RH
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PB_RH
!
!*       0.2  declaration of local variables
!
INTEGER, PARAMETER :: IRCDRYG=1, IRIDRYG=2, IRIWETG=3, IRSDRYG=4, IRSWETG=5, IRRDRYG=6, &
                    & IFREEZ1=7, IFREEZ2=8
!
INTEGER                            :: IGDRY
INTEGER                            :: ISIZE
INTEGER                            :: IDX, JJ, JL
INTEGER, DIMENSION(:), POINTER, CONTIGUOUS :: I1
INTEGER, DIMENSION(:), POINTER, CONTIGUOUS :: IVEC1, IVEC2
LOGICAL, DIMENSION(:), POINTER, CONTIGUOUS :: GDRY
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZDRY, ZDRYG, ZMASK
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZVEC1, ZVEC2, ZVEC3
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZZW,         &
                                              ZRDRYG_INIT, & !Initial dry growth rate of the graupeln
                                              ZRWETG_INIT    !Initial wet growth rate of the graupeln
!
!-------------------------------------------------------------------------------
!
!$acc data present(PCOMPUTE,PRHODREF,PLVFACT,PLSFACT,PPRES,PDV,PKA,PCJ,PCIT,PLBDAR,PLBDAS,PLBDAG,  &
!$acc&                PT,PRVT,PRCT,PRRT,PRIT,PRST,PRGT,PRGSI,PRGSI_MR,PWETG,PRICFRRG,PRRCFRIG,PRICFRR,PRCWETG,PRIWETG, &
!$acc&                PRRWETG,PRSWETG,PRCDRYG,PRIDRYG,PRRDRYG,PRSDRYG,PRWETGH,PRWETGH_MR,PRGMLTR,PRG_TEND,              &
!$acc&                PA_TH,PA_RC,PA_RR,PA_RI,PA_RS,PA_RG,PA_RH,PB_RG,PB_RH)

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PCOMPUTE,"ICE4_FAST_RG beg:PCOMPUTE")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_FAST_RG beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_FAST_RG beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PLSFACT,"ICE4_FAST_RG beg:PLSFACT",PRECISION)
  CALL MPPDB_CHECK1D(PPRES,"ICE4_FAST_RG beg:PPRES",PRECISION)
  CALL MPPDB_CHECK1D(PDV,"ICE4_FAST_RG beg:PDV",PRECISION)
  CALL MPPDB_CHECK1D(PKA,"ICE4_FAST_RG beg:PKA",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_FAST_RG beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PCIT,"ICE4_FAST_RG beg:PCIT",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR,"ICE4_FAST_RG beg:PLBDAR",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAS,"ICE4_FAST_RG beg:PLBDAS",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAG,"ICE4_FAST_RG beg:PLBDAG",PRECISION)
  CALL MPPDB_CHECK1D(PT,"ICE4_FAST_RG beg:PT",PRECISION)
  CALL MPPDB_CHECK1D(PRVT,"ICE4_FAST_RG beg:PRVT",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_FAST_RG beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRRT,"ICE4_FAST_RG beg:PRRT",PRECISION)
  CALL MPPDB_CHECK1D(PRIT,"ICE4_FAST_RG beg:PRIT",PRECISION)
  CALL MPPDB_CHECK1D(PRST,"ICE4_FAST_RG beg:PRST",PRECISION)
  CALL MPPDB_CHECK1D(PRGT,"ICE4_FAST_RG beg:PRGT",PRECISION)
  CALL MPPDB_CHECK1D(PRGSI,"ICE4_FAST_RG beg:PRGSI",PRECISION)
  CALL MPPDB_CHECK1D(PRGSI_MR,"ICE4_FAST_RG beg:PRGSI_MR",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRICFRRG,"ICE4_FAST_RG beg:PRICFRRG",PRECISION)
  CALL MPPDB_CHECK1D(PRRCFRIG,"ICE4_FAST_RG beg:PRRCFRIG",PRECISION)
  CALL MPPDB_CHECK1D(PRICFRR,"ICE4_FAST_RG beg:PRICFRR",PRECISION)
  CALL MPPDB_CHECK1D(PRGMLTR,"ICE4_FAST_RG beg:PRGMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRG_TEND,"ICE4_FAST_RG beg:PRG_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RG beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RG beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RG beg:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RG beg:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RG beg:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RG beg:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RG beg:PA_RH",PRECISION)
  CALL MPPDB_CHECK1D(PB_RG,"ICE4_FAST_RG beg:PB_RG",PRECISION)
  CALL MPPDB_CHECK1D(PB_RH,"ICE4_FAST_RG beg:PB_RH",PRECISION)
END IF

ISIZE = Size( PRHODREF )

#ifndef MNH_OPENACC
allocate( i1         ( isize ) )
allocate( ivec1      ( isize ) )
allocate( ivec2      ( isize ) )
allocate( gdry       ( isize ) )
allocate( zdry       ( isize ) )
allocate( zdryg      ( isize ) )
allocate( zmask      ( isize ) )
allocate( zvec1      ( isize ) )
allocate( zvec2      ( isize ) )
allocate( zvec3      ( isize ) )
allocate( zzw        ( isize ) )
allocate( zrdryg_init( isize ) )
allocate( zrwetg_init( isize ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( I1,    ISIZE )
CALL MNH_MEM_GET( IVEC1, ISIZE )
CALL MNH_MEM_GET( IVEC2, ISIZE )

CALL MNH_MEM_GET( gdry,        isize )

CALL MNH_MEM_GET( zdry,        isize )
CALL MNH_MEM_GET( zdryg,       isize )
CALL MNH_MEM_GET( zmask,       isize )
CALL MNH_MEM_GET( zvec1,       isize )
CALL MNH_MEM_GET( zvec2,       isize )
CALL MNH_MEM_GET( zvec3,       isize )
CALL MNH_MEM_GET( zzw,         isize )
CALL MNH_MEM_GET( zrdryg_init, isize )
CALL MNH_MEM_GET( zrwetg_init, isize )
#endif

!$acc data create ( IGDRY ) &
!$acc&     present( I1, IVEC1, IVEC2, GDRY, ZDRY, ZDRYG, ZMASK, ZVEC1, ZVEC2, ZVEC3, ZZW, ZRDRYG_INIT, ZRWETG_INIT ) &
!$acc&     copyin ( XRTMIN )

!
!*       6.1    rain contact freezing
!
!$acc kernels present_cr( GDRY, PCOMPUTE, PRGT, PRST, XRTMIN, ZDRY )
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(4)-PRIT(JL))) * & ! WHERE(PRIT(:)>XRTMIN(4))
           &MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! WHERE(PRRT(:)>XRTMIN(3))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRICFRRG(JL)=ZMASK(JL) * PRICFRRG(JL)
    PRRCFRIG(JL)=ZMASK(JL) * PRRCFRIG(JL)
    PRICFRR(JL)=ZMASK(JL) * PRICFRR(JL)
  ENDDO
ELSE
  PRICFRRG(:)=0.
  PRRCFRIG(:)=0.
  WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
    PRICFRRG(:) = XICFRR*PRIT(:)                & ! RICFRRG
                                 *PLBDAR(:)**XEXICFRR    &
                                 *PRHODREF(:)**(-XCEXVT)
    PRRCFRIG(:) = XRCFRI*PCIT(:)                & ! RRCFRIG
                                 * PLBDAR(:)**XEXRCFRI    &
                                 * PRHODREF(:)**(-XCEXVT-1.)
#else
    PRICFRRG(:) = XICFRR*PRIT(:)                & ! RICFRRG
                                 *BR_POW(PLBDAR(:),XEXICFRR)    &
                                 *BR_POW(PRHODREF(:),-XCEXVT)
    PRRCFRIG(:) = XRCFRI*PCIT(:)                & ! RRCFRIG
                                 * BR_POW(PLBDAR(:),XEXRCFRI)    &
                                 * BR_POW(PRHODREF(:),-XCEXVT-1.)
#endif
  END WHERE

  IF(LCRFLIMIT) THEN
!$acc loop independent
    DO JL=1, ISIZE
      !Comparison between heat to be released (to freeze rain) and heat sink (rain and ice temperature change)
      !ZZW is the proportion of process that can take place
      ZZW(JL)=(1.-ZMASK(JL)) + & ! 1. outside of mask
              ZMASK(JL) * MAX(0., MIN(1., (PRICFRRG(JL)*XCI+PRRCFRIG(JL)*XCL)*(XTT-PT(JL)) / &
                                          MAX(1.E-20, XLVTT*PRRCFRIG(JL))))
    ENDDO
  ELSE
    ZZW(:)=1.
  ENDIF
  DO JL=1, ISIZE
    PRRCFRIG(JL) = ZZW(JL) * PRRCFRIG(JL) !Part of rain that can be freezed
    PRICFRR(JL) = (1.-ZZW(JL)) * PRICFRRG(JL) !Part of collected pristine ice converted to rain
    PRICFRRG(JL) = ZZW(JL) * PRICFRRG(JL) !Part of collected pristine ice that lead to graupel
  ENDDO
ENDIF
DO JL=1, ISIZE
  PA_RI(JL) = PA_RI(JL) - PRICFRRG(JL) - PRICFRR(JL)
  PA_RR(JL) = PA_RR(JL) - PRRCFRIG(JL) + PRICFRR(JL)
  PA_RG(JL) = PA_RG(JL) + PRICFRRG(JL) + PRRCFRIG(JL)
  PA_TH(JL) = PA_TH(JL) + (PRRCFRIG(JL) - PRICFRR(JL))*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!
!
!*       6.3    compute the graupel growth
!
! Wet and dry collection of rc and ri on graupel
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JL))) * & ! WHERE(PRGT(:)>XRTMIN(6))
           &MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! WHERE(PRCT(:)>XRTMIN(2))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRG_TEND(JL, IRCDRYG)=ZMASK(JL)*PRG_TEND(JL, IRCDRYG)
  ENDDO
ELSE
!$acc loop independent
  DO JL = 1, ISIZE
    IF ( ZMASK(JL) == 1. ) THEN
#ifndef MNH_BITREP
      ZZW(JL) = PLBDAG(JL)**(XCXG-XDG-2.) * PRHODREF(JL)**(-XCEXVT)
#else
      ZZW(JL) = BR_POW( PLBDAG(JL), XCXG - XDG - 2. ) * BR_POW( PRHODREF(JL), -XCEXVT )
#endif
      PRG_TEND(JL, IRCDRYG) = XFCDRYG * PRCT(JL) * ZZW(JL)
    ELSE
      PRG_TEND(JL, IRCDRYG) = 0.
    END IF
  END DO
ENDIF

DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JL))) * & ! WHERE(PRGT(:)>XRTMIN(6))
           &MAX(0., -SIGN(1., XRTMIN(4)-PRIT(JL))) * & ! WHERE(PRIT(:)>XRTMIN(4))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRG_TEND(JL, IRIDRYG)=ZMASK(JL) * PRG_TEND(JL, IRIDRYG)
    PRG_TEND(JL, IRIWETG)=ZMASK(JL) * PRG_TEND(JL, IRIWETG)
  ENDDO
ELSE
!$acc loop independent
  DO JL = 1, ISIZE
    IF ( ZMASK(JL) == 1. ) THEN
#ifndef MNH_BITREP
      ZZW(JL) = PLBDAG(JL)**(XCXG-XDG-2.) * PRHODREF(JL)**(-XCEXVT)
      PRG_TEND(JL, IRIDRYG) = XFIDRYG * EXP( XCOLEXIG * ( PT(JL) - XTT ) ) * PRIT(JL) * ZZW(JL)
      PRG_TEND(JL, IRIWETG) = PRG_TEND(JL, IRIDRYG) / ( XCOLIG * EXP( XCOLEXIG * ( PT(JL) - XTT ) ) )
#else
      ZZW(JL) = BR_POW( PLBDAG(JL), XCXG - XDG - 2. ) * BR_POW( PRHODREF(JL), -XCEXVT )
      PRG_TEND(JL, IRIDRYG) = XFIDRYG * BR_EXP( XCOLEXIG * ( PT(JL) - XTT ) ) * PRIT(JL) * ZZW(JL)
      PRG_TEND(JL, IRIWETG) = PRG_TEND(JL, IRIDRYG) / ( XCOLIG * BR_EXP( XCOLEXIG * ( PT(JL) - XTT ) ) )
#endif
    ELSE
      PRG_TEND(JL, IRIDRYG) = 0.
      PRG_TEND(JL, IRIWETG) = 0.
    END IF
  END DO
ENDIF


! !$acc end kernels
! call mppdb_check(PRG_TEND,"ICE4_FAST_RG:PRG_TEND 02")
! !$acc kernels



! Wet and dry collection of rs on graupel (6.2.1)
IGDRY = 0
!$acc end kernels
!$acc parallel loop private(idx) independent
DO CONCURRENT( JJ = 1 : SIZE( GDRY ) )
  ZDRY(JJ)=MAX(0., -SIGN(1., XRTMIN(5)-PRST(JJ))) * & ! WHERE(PRST(:)>XRTMIN(5))
          &MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JJ))) * & ! WHERE(PRGT(:)>XRTMIN(6))
          &PCOMPUTE(JJ)
  IF (ZDRY(JJ)>0) THEN
!$acc atomic capture
    IGDRY = IGDRY + 1
    IDX = IGDRY
!$acc end atomic
    I1(IDX) = JJ
    GDRY(JJ) = .TRUE.
  ELSE
    GDRY(JJ) = .FALSE.
  END IF
END DO
! !$acc kernels
IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRG_TEND(JL, IRSDRYG)=ZDRY(JL) * PRG_TEND(JL, IRSDRYG)
    PRG_TEND(JL, IRSWETG)=ZDRY(JL) * PRG_TEND(JL, IRSWETG)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRG_TEND(:, IRSDRYG)=0.
  PRG_TEND(:, IRSWETG)=0.
!$acc end kernels

!$acc update self( IGDRY )
  IF(IGDRY>0)THEN
!$acc kernels
    !
    !*       6.2.3  select the (PLBDAG,PLBDAS) couplet
    !
!$acc loop independent
    DO JJ = 1, IGDRY
      ZVEC1(JJ) = PLBDAG(I1(JJ))
      ZVEC2(JJ) = PLBDAS(I1(JJ))
    END DO
    !
    !*       6.2.4  find the next lower indice for the PLBDAG and for the PLBDAS
    !               in the geometrical set of (Lbda_g,Lbda_s) couplet use to
    !               tabulate the SDRYG-kernel
    !
!$acc loop independent
    DO CONCURRENT( JJ = 1 : IGDRY )
      ZVEC1(JJ) = MAX( 1.00001, MIN( REAL( NDRYLBDAG ) - 0.00001,           &
#ifndef MNH_BITREP
                            XDRYINTP1G * LOG( ZVEC1(JJ) ) + XDRYINTP2G ) )
#else
                            XDRYINTP1G * BR_LOG( ZVEC1(JJ) ) + XDRYINTP2G ) )
#endif
      IVEC1(JJ) = INT( ZVEC1(JJ) )
      ZVEC1(JJ) = ZVEC1(JJ) - REAL( IVEC1(JJ) )
      !
      ZVEC2(JJ) = MAX( 1.00001, MIN( REAL( NDRYLBDAS ) - 0.00001,           &
#ifndef MNH_BITREP
                            XDRYINTP1S * LOG( ZVEC2(JJ) ) + XDRYINTP2S ) )
#else
                            XDRYINTP1S * BR_LOG( ZVEC2(JJ) ) + XDRYINTP2S ) )
#endif
      IVEC2(JJ) = INT( ZVEC2(JJ) )
      ZVEC2(JJ) = ZVEC2(JJ) - REAL( IVEC2(JJ) )
    END DO
    !
    !*       6.2.5  perform the bilinear interpolation of the normalized
    !               SDRYG-kernel
    !
!$acc loop independent
    DO JJ=1, IGDRY
      ZVEC3(JJ) =  (  XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * ZVEC1(JJ) &
                 - (  XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         *(ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGDRY
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GDRY(:))
#ifndef MNH_BITREP
      PRG_TEND(:, IRSWETG)=XFSDRYG*ZZW(:)                         & ! RSDRYG
                                    / XCOLSG &
                  *(PLBDAS(:)**(XCXS-XBS))*( PLBDAG(:)**XCXG )    &
                  *(PRHODREF(:)**(-XCEXVT-1.))                    &
                       *( XLBSDRYG1/( PLBDAG(:)**2              ) + &
                          XLBSDRYG2/( PLBDAG(:)   * PLBDAS(:)   ) + &
                          XLBSDRYG3/(               PLBDAS(:)**2))
      PRG_TEND(:, IRSDRYG)=PRG_TEND(:, IRSWETG)*XCOLSG*EXP(XCOLEXSG*(PT(:)-XTT))
#else
      PRG_TEND(:, IRSWETG)=XFSDRYG*ZZW(:)                            & ! RSDRYG
                                    / XCOLSG                         &
                  *BR_POW(PLBDAS(:),XCXS-XBS)*BR_POW(PLBDAG(:),XCXG) &
                  *BR_POW(PRHODREF(:),-XCEXVT-1.)                    &
                       *( XLBSDRYG1/  BR_P2(PLBDAG(:))             + &
                          XLBSDRYG2/( PLBDAG(:)   * PLBDAS(:)   )  + &
                          XLBSDRYG3/              BR_P2(PLBDAS(:)) )
      PRG_TEND(:, IRSDRYG)=PRG_TEND(:, IRSWETG)*XCOLSG*BR_EXP(XCOLEXSG*(PT(:)-XTT))
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!acc kernels present_cr( GDRY, PCOMPUTE, PRGT, PRRT, XRTMIN, ZDRY )
!
!*       6.2.6  accretion of raindrops on the graupeln
!
IGDRY = 0
!$acc parallel loop private(idx) independent
DO CONCURRENT( JJ = 1 : SIZE( GDRY ) )
  ZDRY(JJ)=MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JJ))) * & ! WHERE(PRRT(:)>XRTMIN(3))
          &MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JJ))) * & ! WHERE(PRGT(:)>XRTMIN(6))
          &PCOMPUTE(JJ)
  IF (ZDRY(JJ)>0) THEN
!$acc atomic capture
    IGDRY = IGDRY + 1
    IDX = IGDRY
!$acc end atomic
    I1(IDX) = JJ
    GDRY(JJ) = .TRUE.
  ELSE
    GDRY(JJ) = .FALSE.
  END IF
END DO

IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRG_TEND(JL, IRRDRYG)=ZDRY(JL) * PRG_TEND(JL, IRRDRYG)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRG_TEND(:, IRRDRYG)=0.
!$acc end kernels
!$acc update self( IGDRY )
  !
!$acc kernels
  IF(IGDRY>0) THEN
    !
    !*       6.2.8  select the (PLBDAG,PLBDAR) couplet
    !
!$acc loop independent
    DO JJ = 1, IGDRY
      ZVEC1(JJ) = PLBDAG(I1(JJ))
      ZVEC2(JJ) = PLBDAR(I1(JJ))
    END DO
    !
    !*       6.2.9  find the next lower indice for the PLBDAG and for the PLBDAR
    !               in the geometrical set of (Lbda_g,Lbda_r) couplet use to
    !               tabulate the RDRYG-kernel
    !
!$acc loop independent
    DO JJ = 1, IGDRY
      ZVEC1(JJ) = MAX( 1.00001, MIN( REAL( NDRYLBDAG ) - 0.00001,           &
#ifndef MNH_BITREP
                            XDRYINTP1G * LOG( ZVEC1(JJ) ) + XDRYINTP2G ) )
#else
                            XDRYINTP1G * BR_LOG( ZVEC1(JJ) ) + XDRYINTP2G ) )
#endif
      IVEC1(JJ) = INT( ZVEC1(JJ) )
      ZVEC1(JJ) = ZVEC1(JJ) - REAL( IVEC1(JJ) )
      !
      ZVEC2(JJ) = MAX( 1.00001, MIN( REAL( NDRYLBDAR ) - 0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1R * LOG( ZVEC2(JJ) ) + XDRYINTP2R ) )
#else
                          XDRYINTP1R * BR_LOG( ZVEC2(JJ) ) + XDRYINTP2R ) )
#endif
      IVEC2(JJ) = INT( ZVEC2(JJ) )
      ZVEC2(JJ) = ZVEC2(JJ) - REAL( IVEC2(JJ) )
    END DO
    !
    !*       6.2.10 perform the bilinear interpolation of the normalized
    !               RDRYG-kernel
    !
!$acc loop independent
    DO JJ=1, IGDRY
      ZVEC3(JJ)= (  XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                  * ZVEC1(JJ) &
                 - (  XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         *(ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGDRY
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GDRY(:))
#ifndef MNH_BITREP
      PRG_TEND(:, IRRDRYG) = XFRDRYG*ZZW(:)                    & ! RRDRYG
                        *( PLBDAR(:)**(-4) )*( PLBDAG(:)**XCXG ) &
                               *( PRHODREF(:)**(-XCEXVT-1.) )   &
                    *( XLBRDRYG1/( PLBDAG(:)**2              ) + &
                       XLBRDRYG2/( PLBDAG(:)   * PLBDAR(:)   ) + &
                       XLBRDRYG3/(               PLBDAR(:)**2) )
#else
      PRG_TEND(:, IRRDRYG) = XFRDRYG*ZZW(:)                         & ! RRDRYG
                        / BR_P4(PLBDAR(:)) * BR_POW(PLBDAG(:),XCXG) &
                               * BR_POW(PRHODREF(:),-XCEXVT-1.)     &
                    *( XLBRDRYG1/ BR_P2(PLBDAG(:))                + &
                       XLBRDRYG2/( PLBDAG(:)   * PLBDAR(:)   )    + &
                       XLBRDRYG3/               BR_P2(PLBDAR(:))  )
#endif
    END WHERE
  ENDIF
!$acc end kernels
ENDIF
!$acc kernels

DO JL=1, ISIZE
  ZRDRYG_INIT(JL)=PRG_TEND(JL, IRCDRYG)+PRG_TEND(JL, IRIDRYG)+ &
                 &PRG_TEND(JL, IRSDRYG)+PRG_TEND(JL, IRRDRYG)
ENDDO

!Freezing rate
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JL))) * & ! WHERE(PRGT(:)>XRTMIN(6))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRG_TEND(JL, IFREEZ1)=ZMASK(JL) * PRG_TEND(JL, IFREEZ1)
    PRG_TEND(JL, IFREEZ2)=ZMASK(JL) * PRG_TEND(JL, IFREEZ2)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRG_TEND(JL, IFREEZ1)=ZMASK(JL) * PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRG_TEND(:, IFREEZ1)=MIN(PRG_TEND(:, IFREEZ1), EXP(XALPI-XBETAI/PT(:)-XGAMI*ALOG(PT(:)))) ! min(ev, es_i(T))
#else
      PRG_TEND(:, IFREEZ1)=MIN(PRG_TEND(:, IFREEZ1), BR_EXP(XALPI-XBETAI/PT(:)-XGAMI*BR_LOG(PT(:)))) ! min(ev, es_i(T))
#endif
    END WHERE
  ENDIF
  PRG_TEND(:, IFREEZ2)=0.
  WHERE(ZMASK(:)==1.)
    PRG_TEND(:, IFREEZ1)=PKA(:)*(XTT-PT(:)) +                              &
             (PDV(:)*(XLVTT+(XCPV-XCL)*(PT(:)-XTT)) &
                           *(XESTT-PRG_TEND(:, IFREEZ1))/(XRV*PT(:))           )
#ifndef MNH_BITREP
    PRG_TEND(:, IFREEZ1)=PRG_TEND(:, IFREEZ1)* ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                           X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG )/ &
#else
    PRG_TEND(:, IFREEZ1)=PRG_TEND(:, IFREEZ1)* ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                           X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) )/ &
#endif
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
    PRG_TEND(:, IFREEZ2)=(PRHODREF(:)*(XLMTT+(XCI-XCL)*(XTT-PT(:)))   ) / &
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
  END WHERE
ENDIF
!$acc loop independent
DO JL=1, ISIZE
  !We must agregate, at least, the cold species
  ZRWETG_INIT(JL)=ZMASK(JL) * MAX(PRG_TEND(JL, IRIWETG)+PRG_TEND(JL, IRSWETG), &
                                 &MAX(0., PRG_TEND(JL, IFREEZ1) + &
                                         &PRG_TEND(JL, IFREEZ2) * ( &
                     &PRG_TEND(JL, IRIWETG)+PRG_TEND(JL, IRSWETG) )))
ENDDO

!Growth mode
DO JL=1, ISIZE
  PWETG(JL) = ZMASK(JL) * & !
            & MAX(0., SIGN(1., MAX(0., ZRDRYG_INIT(JL)-PRG_TEND(JL, IRIDRYG)-PRG_TEND(JL, IRSDRYG)) - &
                              &MAX(0., ZRWETG_INIT(JL)-PRG_TEND(JL, IRIWETG)-PRG_TEND(JL, IRSWETG))))
ENDDO
IF(LNULLWETG) THEN
  DO JL=1, ISIZE
    PWETG(JL) = PWETG(JL) * MAX(0., -SIGN(1., -ZRDRYG_INIT(JL)))
  ENDDO
ELSE
  DO JL=1, ISIZE
    PWETG(JL) = PWETG(JL) * MAX(0., -SIGN(1., -ZRWETG_INIT(JL)))
  ENDDO
ENDIF
IF(.NOT. LWETGPOST) THEN
  DO JL=1, ISIZE
    PWETG(JL) = PWETG(JL) * MAX(0., -SIGN(1., PT(JL)-XTT))
  ENDDO
ENDIF
!$acc loop independent
DO JL=1, ISIZE
  ZDRYG(JL) = ZMASK(JL) * & !
            & MAX(0., -SIGN(1., PT(JL)-XTT)) * & ! WHERE(PT(:)<XTT)
            & MAX(0., -SIGN(1., 1.E-20-ZRDRYG_INIT(JL))) * & ! WHERE(ZRDRYG_INIT(:)>0.)
            & MAX(0., -SIGN(1., MAX(0., ZRDRYG_INIT(JL)-PRG_TEND(JL, IRIDRYG)-PRG_TEND(JL, IRSDRYG)) - &
                               &MAX(0., ZRWETG_INIT(JL)-PRG_TEND(JL, IRIWETG)-PRG_TEND(JL, IRSWETG))))
ENDDO

! Part of ZRWETG to be converted into hail
! Graupel can be produced by other processes instantaneously (inducing a mixing ratio change, PRGSI_MR) or
! as a tendency (PRWETGH)
PRWETGH(:)=0.
PRWETGH_MR(:)=0.
IF(KRR==7) THEN
  WHERE(PWETG(:)==1.)
    !assume a linear percent of conversion of produced graupel into hail
    PRWETGH(:)=(MAX(0., PRGSI(:)+PRICFRRG(:)+PRRCFRIG(:))+ZRWETG_INIT(:))*ZRDRYG_INIT(:)/(ZRWETG_INIT(:)+ZRDRYG_INIT(:))
    PRWETGH_MR(:)=MAX(0., PRGSI_MR(:))*ZRDRYG_INIT(:)/(ZRWETG_INIT(:)+ZRDRYG_INIT(:))
  END WHERE
ENDIF

DO JL=1, ISIZE
  !Aggregated minus collected
  PRRWETG(JL)=-PWETG(JL) * (PRG_TEND(JL, IRIWETG)+PRG_TEND(JL, IRSWETG)+&
                           &PRG_TEND(JL, IRCDRYG)-ZRWETG_INIT(JL))
  PRCWETG(JL)=PWETG(JL) * PRG_TEND(JL, IRCDRYG)
  PRIWETG(JL)=PWETG(JL) * PRG_TEND(JL, IRIWETG)
  PRSWETG(JL)=PWETG(JL) * PRG_TEND(JL, IRSWETG)

  PRCDRYG(JL)=ZDRYG(JL) * PRG_TEND(JL, IRCDRYG)
  PRRDRYG(JL)=ZDRYG(JL) * PRG_TEND(JL, IRRDRYG)
  PRIDRYG(JL)=ZDRYG(JL) * PRG_TEND(JL, IRIDRYG)
  PRSDRYG(JL)=ZDRYG(JL) * PRG_TEND(JL, IRSDRYG)

  PA_RC(JL) = PA_RC(JL) - PRCWETG(JL)
  PA_RI(JL) = PA_RI(JL) - PRIWETG(JL)
  PA_RS(JL) = PA_RS(JL) - PRSWETG(JL)
  PA_RG(JL) = PA_RG(JL) + PRCWETG(JL) + PRIWETG(JL) + PRSWETG(JL) + PRRWETG(JL)
  PA_RR(JL) = PA_RR(JL) - PRRWETG(JL)
  PA_TH(JL) = PA_TH(JL) + (PRCWETG(JL) + PRRWETG(JL))*(PLSFACT(JL)-PLVFACT(JL))
  PA_RG(JL) = PA_RG(JL) - PRWETGH(JL)
  PA_RH(JL) = PA_RH(JL) + PRWETGH(JL)
  PB_RG(JL) = PB_RG(JL) - PRWETGH_MR(JL)
  PB_RH(JL) = PB_RH(JL) + PRWETGH_MR(JL)
  PA_RC(JL) = PA_RC(JL) - PRCDRYG(JL)
  PA_RI(JL) = PA_RI(JL) - PRIDRYG(JL)
  PA_RS(JL) = PA_RS(JL) - PRSDRYG(JL)
  PA_RR(JL) = PA_RR(JL) - PRRDRYG(JL)
  PA_RG(JL) = PA_RG(JL) + PRCDRYG(JL) + PRIDRYG(JL) + PRSDRYG(JL) + PRRDRYG(JL)
  PA_TH(JL) = PA_TH(JL) + (PRCDRYG(JL)+PRRDRYG(JL))*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!
!*       6.5    Melting of the graupeln
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JL))) * & ! WHERE(PRGT(:)>XRTMIN(6))
           &MAX(0., -SIGN(1., XTT-PT(JL))) * & ! WHERE(PT(:)>XTT)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRGMLTR(JL)=ZMASK(JL) * PRGMLTR(JL)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRGMLTR(JL)=ZMASK(JL) * PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRGMLTR(:)=MIN(PRGMLTR(:), EXP(XALPW-XBETAW/PT(:)-XGAMW*ALOG(PT(:)))) ! min(ev, es_w(T))
#else
      PRGMLTR(:)=MIN(PRGMLTR(:), BR_EXP(XALPW-XBETAW/PT(:)-XGAMW*BR_LOG(PT(:)))) ! min(ev, es_w(T))
#endif
    END WHERE
  ENDIF
  DO JL=1, ISIZE
    PRGMLTR(JL)=ZMASK(JL) * (PKA(JL)*(XTT-PT(JL)) +                                 &
               ( PDV(JL)*(XLVTT + ( XCPV - XCL ) * ( PT(JL) - XTT )) &
                           *(XESTT-PRGMLTR(JL))/(XRV*PT(JL))             ))
  ENDDO
  WHERE(ZMASK(:)==1.)
    !
    ! compute RGMLTR
    !
    PRGMLTR(:)  = MAX( 0.0,( -PRGMLTR(:) *                     &
#ifndef MNH_BITREP
                           ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                             X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG ) -   &
#else
                           ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                             X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) ) -   &
#endif
                         ( PRG_TEND(:, IRCDRYG)+PRG_TEND(:, IRRDRYG) ) *       &
                               ( PRHODREF(:)*XCL*(XTT-PT(:))) ) /    &
                                             ( PRHODREF(:)*XLMTT ) )
  END WHERE
ENDIF
DO JL=1, ISIZE
  PA_RR(JL) = PA_RR(JL) + PRGMLTR(JL)
  PA_RG(JL) = PA_RG(JL) - PRGMLTR(JL)
  PA_TH(JL) = PA_TH(JL) - PRGMLTR(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRICFRRG,"ICE4_FAST_RG end:PRICFRRG",PRECISION)
  CALL MPPDB_CHECK1D(PRRCFRIG,"ICE4_FAST_RG end:PRRCFRIG",PRECISION)
  CALL MPPDB_CHECK1D(PRICFRR,"ICE4_FAST_RG end:PRICFRR",PRECISION)
  CALL MPPDB_CHECK1D(PRGMLTR,"ICE4_FAST_RG end:PRGMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRG_TEND,"ICE4_FAST_RG end:PRG_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RG end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RG end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RG end:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RG end:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RG end:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RG end:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RG end:PA_RH",PRECISION)
  CALL MPPDB_CHECK1D(PB_RG,"ICE4_FAST_RG end:PB_RG",PRECISION)
  CALL MPPDB_CHECK1D(PB_RH,"ICE4_FAST_RG end:PB_RH",PRECISION)
  !Check all OUT arrays
  CALL MPPDB_CHECK1D(PWETG,"ICE4_FAST_RG end:PWETG")
  CALL MPPDB_CHECK1D(PRCWETG,"ICE4_FAST_RG end:PRCWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRIWETG,"ICE4_FAST_RG end:PRIWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRRWETG,"ICE4_FAST_RG end:PRRWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRSWETG,"ICE4_FAST_RG end:PRSWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRCDRYG,"ICE4_FAST_RG end:PRCDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRIDRYG,"ICE4_FAST_RG end:PRIDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRRDRYG,"ICE4_FAST_RG end:PRRDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRSDRYG,"ICE4_FAST_RG end:PRSDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRWETGH,"ICE4_FAST_RG end:PRWETGH",PRECISION)
  CALL MPPDB_CHECK1D(PRWETGH_MR,"ICE4_FAST_RG end:PRWETGH_MR",PRECISION)
END IF

!$acc end data

#ifndef MNH_OPENACC
deallocate( zrwetg_init )
deallocate( zrdryg_init )
deallocate( zzw         )
deallocate( zvec3       )
deallocate( zvec2       )
deallocate( zvec1       )
deallocate( zmask       )
deallocate( zdryg       )
deallocate( zdry        )
deallocate( gdry        )
deallocate( ivec2       )
deallocate( ivec1       )
deallocate( i1          )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data

END SUBROUTINE ICE4_FAST_RG
