!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_SEDIMENTATION_SPLIT
INTERFACE
SUBROUTINE ICE4_SEDIMENTATION_SPLIT(KIB, KIE, KIT, KJB, KJE, KJT, KKB, KKE, KKTB, KKTE, KKT, KKL, &
                                   &PTSTEP, KRR, OSEDIC, ODEPOSC, PVDEPOSC, PDZZ, &
                                   &PRHODREF, PPABST, PTHT, PRHODJ, &
                                   &PRCS, PRCT, PRRS, PRRT, PRIS, PRIT, PRSS, PRST, PRGS, PRGT,&
                                   &PINPRC, PINDEP, PINPRR, PINPRI, PINPRS, PINPRG, &
                                   &PSEA, PTOWN,  &
                                   &PINPRH, PRHT, PRHS, PFPR)
IMPLICIT NONE
INTEGER,                            INTENT(IN)    :: KIB, KIE, KIT, KJB, KJE, KJT, KKB, KKE, KKTB, KKTE, KKT
INTEGER,                            INTENT(IN)    :: KKL     !vert. levels type 1=MNH -1=ARO
REAL,                               INTENT(IN)    :: PTSTEP  ! Double Time step (single if cold start)
INTEGER,                            INTENT(IN)    :: KRR     ! Number of moist variable
LOGICAL,                            INTENT(IN)    :: OSEDIC  ! Switch for droplet sedim.
LOGICAL,                            INTENT(IN)    :: ODEPOSC ! Switch for droplet depos.
REAL,                               INTENT(IN)    :: PVDEPOSC! Droplet deposition velocity
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PDZZ    ! Layer thikness (m)
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PTHT    ! Theta at time t
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRHODJ  ! Dry density * Jacobian
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRCS    ! Cloud water m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRCT    ! Cloud water m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRRS    ! Rain water m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRRT    ! Rain water m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRIS    ! Pristine ice m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRIT    ! Pristine ice m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRSS    ! Snow/aggregate m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRST    ! Snow/aggregate m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRGS    ! Graupel m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRGT    ! Graupel/hail m.r. at t
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRC  ! Cloud instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINDEP  ! Cloud instant deposition
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRR  ! Rain instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRI  ! Pristine ice instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRS  ! Snow instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRG  ! Graupel instant precip
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(IN)    :: PSEA    ! Sea Mask
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(IN)    :: PTOWN   ! Fraction that is town
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(OUT)   :: PINPRH  ! Hail instant precip
REAL, DIMENSION(:,:,:),   OPTIONAL, INTENT(IN)    :: PRHT    ! Hail m.r. at t
REAL, DIMENSION(:,:,:),   OPTIONAL, INTENT(INOUT) :: PRHS    ! Hail m.r. source
REAL, DIMENSION(:,:,:,:), OPTIONAL, INTENT(OUT)   :: PFPR    ! upper-air precipitation fluxes
END SUBROUTINE ICE4_SEDIMENTATION_SPLIT
END INTERFACE
END MODULE MODI_ICE4_SEDIMENTATION_SPLIT
SUBROUTINE ICE4_SEDIMENTATION_SPLIT(KIB, KIE, KIT, KJB, KJE, KJT, KKB, KKE, KKTB, KKTE, KKT, KKL, &
                                   &PTSTEP, KRR, OSEDIC, ODEPOSC, PVDEPOSC, PDZZ, &
                                   &PRHODREF, PPABST, PTHT, PRHODJ, &
                                   &PRCS, PRCT, PRRS, PRRT, PRIS, PRIT, PRSS, PRST, PRGS, PRGT,&
                                   &PINPRC, PINDEP, PINPRR, PINPRI, PINPRS, PINPRG, &
                                   &PSEA, PTOWN,  &
                                   &PINPRH, PRHT, PRHS, PFPR)
!!
!!**  PURPOSE
!!    -------
!!      Computes the sedimentation
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the plitting of rain_ice source code (nov. 2014)
!!                and modified for optimisation
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 11/02/2019: dimensions of PINPRC and PINDEP not necessarily KIT,KJT
!  P. Wautelet 10/04/2019: replace ABORT and STOP calls by Print_msg
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XRHOLW
USE MODD_PARAM_ICE,      ONLY: XSPLIT_MAXCFL
USE MODD_RAIN_ICE_DESCR, ONLY: XALPHAC,XALPHAC2,XCONC_LAND,XCONC_SEA,XCONC_URBAN,XLBC,XNUC,XNUC2
USE MODD_RAIN_ICE_PARAM, ONLY: XFSEDC
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
!
#ifndef MNH_COMPILER_CCE
!Do not use internal GAMMA function but the Fortran intrinsics one with Cray compiler
!Warning: intrinsics gamma does not give same results between CPU and GPU with NVHPC (tested with 22.2 version)
USE MODI_GAMMA
#endif
#if defined(TARGET_NV70)
USE MODI_GAMMA
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                            INTENT(IN)    :: KIB, KIE, KIT, KJB, KJE, KJT, KKB, KKE, KKTB, KKTE, KKT
INTEGER,                            INTENT(IN)    :: KKL     !vert. levels type 1=MNH -1=ARO
REAL,                               INTENT(IN)    :: PTSTEP  ! Double Time step (single if cold start)
INTEGER,                            INTENT(IN)    :: KRR     ! Number of moist variable
LOGICAL,                            INTENT(IN)    :: OSEDIC  ! Switch for droplet sedim.
LOGICAL,                            INTENT(IN)    :: ODEPOSC ! Switch for droplet depos.
REAL,                               INTENT(IN)    :: PVDEPOSC! Droplet deposition velocity
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PDZZ    ! Layer thikness (m)
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PTHT    ! Theta at time t
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRHODJ  ! Dry density * Jacobian
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRCS    ! Cloud water m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRCT    ! Cloud water m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRRS    ! Rain water m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRRT    ! Rain water m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRIS    ! Pristine ice m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRIT    ! Pristine ice m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRSS    ! Snow/aggregate m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRST    ! Snow/aggregate m.r. at t
REAL, DIMENSION(:,:,:),             INTENT(INOUT) :: PRGS    ! Graupel m.r. source
REAL, DIMENSION(:,:,:),             INTENT(IN)    :: PRGT    ! Graupel/hail m.r. at t
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRC  ! Cloud instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINDEP  ! Cloud instant deposition
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRR  ! Rain instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRI  ! Pristine ice instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRS  ! Snow instant precip
REAL, DIMENSION(:,:),               INTENT(OUT)   :: PINPRG  ! Graupel instant precip
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(IN)    :: PSEA    ! Sea Mask
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(IN)    :: PTOWN   ! Fraction that is town
REAL, DIMENSION(:,:),     OPTIONAL, INTENT(OUT)   :: PINPRH  ! Hail instant precip
REAL, DIMENSION(:,:,:),   OPTIONAL, INTENT(IN)    :: PRHT    ! Hail m.r. at t
REAL, DIMENSION(:,:,:),   OPTIONAL, INTENT(INOUT) :: PRHS    ! Hail m.r. source
REAL, DIMENSION(:,:,:,:), OPTIONAL, INTENT(OUT)   :: PFPR    ! upper-air precipitation fluxes
!
!
!*       0.2  declaration of local variables
!
!
INTEGER                             :: JI,JJ,JK
LOGICAL                             :: GPRESENT_PFPR, GPRESENT_PSEA
REAL                                :: ZINVTSTEP
#ifndef MNH_OPENACC
REAL, DIMENSION(:,:),   allocatable :: ZCONC_TMP    ! Weighted concentration
REAL, DIMENSION(:,:,:), allocatable :: ZW ! work array
REAL, DIMENSION(:,:,:), allocatable :: ZCONC3D, & !  droplet condensation
                                     & ZRAY,   & ! Cloud Mean radius
                                     & ZLBC,   & ! XLBC weighted by sea fraction
                                     & ZFSEDC, &
                                     & ZPRCS,ZPRRS,ZPRIS,ZPRSS,ZPRGS,ZPRHS, &   ! Mixing ratios created during the time step
                                     & ZRCT, &
                                     & ZRRT, &
                                     & ZRIT, &
                                     & ZRST, &
                                     & ZRGT, &
                                     & ZRHT
#else
REAL, DIMENSION(:,:),   POINTER, CONTIGUOUS :: ZCONC_TMP    ! Weighted concentration
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZW ! work array
REAL, DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZCONC3D, & !  droplet condensation
                                               ZRAY,   & ! Cloud Mean radius
                                               ZLBC,   & ! XLBC weighted by sea fraction
                                               ZFSEDC, &
                                               ZPRCS,ZPRRS,ZPRIS,ZPRSS,ZPRGS,ZPRHS, & ! Mixing ratios created during the time step
                                               ZRCT, &
                                               ZRRT, &
                                               ZRIT, &
                                               ZRST, &
                                               ZRGT, &
                                               ZRHT
#endif
!-------------------------------------------------------------------------------

!$acc data present( PDZZ, PRHODREF, PPABST, PTHT, PRHODJ, PRCS, PRCT, PRRS, PRRT, PRIS, PRIT, PRSS, PRST, PRGS, PRGT,                 &
!$acc&              PINPRC, PINDEP, PINPRR, PINPRI, PINPRS, PINPRG, PSEA, PTOWN, PINPRH, PRHT, PRHS, PFPR )

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PDZZ,"ICE4_SEDIMENTATION_SPLIT beg:PDZZ")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_SEDIMENTATION_SPLIT beg:PRHODREF")
  CALL MPPDB_CHECK(PPABST,"ICE4_SEDIMENTATION_SPLIT beg:PPABST")
  CALL MPPDB_CHECK(PTHT,"ICE4_SEDIMENTATION_SPLIT beg:PTHT")
  CALL MPPDB_CHECK(PRHODJ,"ICE4_SEDIMENTATION_SPLIT beg:PRHODJ")
  CALL MPPDB_CHECK(PRCT,"ICE4_SEDIMENTATION_SPLIT beg:PRCT")
  CALL MPPDB_CHECK(PRRT,"ICE4_SEDIMENTATION_SPLIT beg:PRRT")
  CALL MPPDB_CHECK(PRIT,"ICE4_SEDIMENTATION_SPLIT beg:PRIT")
  CALL MPPDB_CHECK(PRST,"ICE4_SEDIMENTATION_SPLIT beg:PRST")
  CALL MPPDB_CHECK(PRGT,"ICE4_SEDIMENTATION_SPLIT beg:PRGT")
  IF (PRESENT(PSEA))  CALL MPPDB_CHECK(PSEA,"ICE4_SEDIMENTATION_SPLIT beg:PSEA")
  IF (PRESENT(PTOWN)) CALL MPPDB_CHECK(PTOWN,"ICE4_SEDIMENTATION_SPLIT beg:PTOWN")
  IF (PRESENT(PRHT))  CALL MPPDB_CHECK(PRHT,"ICE4_SEDIMENTATION_SPLIT beg:PRHT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"ICE4_SEDIMENTATION_SPLIT beg:PRCS")
  CALL MPPDB_CHECK(PRRS,"ICE4_SEDIMENTATION_SPLIT beg:PRRS")
  CALL MPPDB_CHECK(PRIS,"ICE4_SEDIMENTATION_SPLIT beg:PRIS")
  CALL MPPDB_CHECK(PRSS,"ICE4_SEDIMENTATION_SPLIT beg:PRSS")
  CALL MPPDB_CHECK(PRGS,"ICE4_SEDIMENTATION_SPLIT beg:PRGS")
  IF (PRESENT(PRHS)) CALL MPPDB_CHECK(PRHS,"ICE4_SEDIMENTATION_SPLIT beg:PRHS")
END IF


#ifndef MNH_OPENACC
allocate( zconc_tmp(size( prhodref, 1 ), size( prhodref, 2 ) ) )

allocate( zw(size( prhodref, 1 ), size( prhodref, 2 ), KKTB : KKTE ) )

allocate( zconc3d(size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zray   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zlbc   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zfsedc (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zprcs  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zprrs  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zpris  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zprss  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zprgs  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zprhs  (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrct   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrrt   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrit   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrst   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrgt   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
allocate( zrht   (size( prhodref, 1 ), size( prhodref, 2 ), size( prhodref, 3 ) ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( zconc_tmp, KIT , KJT )

CALL MNH_MEM_GET( zw, 1, KIT, 1, KJT, KKTB, KKTE )

CALL MNH_MEM_GET( zconc3d, KIT, KJT, KKT )
CALL MNH_MEM_GET( zray   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zlbc   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zfsedc , KIT, KJT, KKT )
CALL MNH_MEM_GET( zprcs  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zprrs  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zpris  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zprss  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zprgs  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zprhs  , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrct   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrrt   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrit   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrst   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrgt   , KIT, KJT, KKT )
CALL MNH_MEM_GET( zrht   , KIT, KJT, KKT )

!$acc data present( zconc_tmp, zw,                                                         &
!$acc&              zconc3d, zray, zlbc, zfsedc, zprcs, zprrs, zpris, zprss, zprgs, zprhs, &
!$acc&              zrct, zrrt, zrit, zrst, zrgt, zrht )

#endif


!
IF (PRESENT(PFPR)) THEN
  GPRESENT_PFPR = .TRUE.
ELSE
  GPRESENT_PFPR = .FALSE.
END IF
!
IF (PRESENT(PSEA)) THEN
  GPRESENT_PSEA = .TRUE.
ELSE
  GPRESENT_PSEA = .FALSE.
END IF
!
!        O. Initialization of for sedimentation
!
!$acc kernels
ZINVTSTEP=1./PTSTEP
!$acc end kernels
IF (GPRESENT_PFPR) THEN
!$acc kernels
  PFPR(:,:,:,:) = 0.
!$acc end kernels
END IF
!
!*       1. Parameters for cloud sedimentation
!
IF (OSEDIC) THEN
!$acc kernels
  ZRAY(:,:,:)   = 0.
  ZLBC(:,:,:)   = XLBC(1)
  ZFSEDC(:,:,:) = XFSEDC(1)
  ZCONC3D(:,:,:)= XCONC_LAND
  ZCONC_TMP(:,:)= XCONC_LAND
!$acc end kernels
  IF (GPRESENT_PSEA) THEN
!$acc kernels
   ZCONC_TMP(:,:)=PSEA(:,:)*XCONC_SEA+(1.-PSEA(:,:))*XCONC_LAND
!$acc end kernels
#if 0
!$acc kernels
!$acc loop independent collapse(3)
    DO JK = KKTB, KKTE
      DO JJ = 1, KJT
        DO JI = 1, KIT
          ZLBC(JI,JJ,JK)   = PSEA(JI,JJ)*XLBC(2)+(1.-PSEA(JI,JJ))*XLBC(1)
          ZFSEDC(JI,JJ,JK) = (PSEA(JI,JJ)*XFSEDC(2)+(1.-PSEA(JI,JJ))*XFSEDC(1))
          ZFSEDC(JI,JJ,JK) = MAX(MIN(XFSEDC(1),XFSEDC(2)),ZFSEDC(JI,JJ,JK))
          ZCONC3D(JI,JJ,JK)= (1.-PTOWN(JI,JJ))*ZCONC_TMP(JI,JJ)+PTOWN(JI,JJ)*XCONC_URBAN
          ZRAY(JI,JJ,JK)   = 0.5*((1.-PSEA(JI,JJ))*GAMMA(XNUC+1.0/XALPHAC)/(GAMMA(XNUC)) + &
                  PSEA(JI,JJ)*GAMMA(XNUC2+1.0/XALPHAC2)/(GAMMA(XNUC2)))
        END DO
      END DO
    END DO
!$acc end kernels
#else
!$acc kernels present_cr( PSEA, PTOWN, ZCONC3D, ZFSEDC, ZLBC, ZRAY )
!$acc loop independent
    DO CONCURRENT( JI = 1 : KIT, JJ = 1 : KJT, JK = KKTB : KKTE )
      ZLBC(JI,JJ,JK)   = PSEA(JI,JJ)*XLBC(2)+(1.-PSEA(JI,JJ))*XLBC(1)
      ZFSEDC(JI,JJ,JK) = (PSEA(JI,JJ)*XFSEDC(2)+(1.-PSEA(JI,JJ))*XFSEDC(1))
      ZFSEDC(JI,JJ,JK) = MAX(MIN(XFSEDC(1),XFSEDC(2)),ZFSEDC(JI,JJ,JK))
      ZCONC3D(JI,JJ,JK)= (1.-PTOWN(JI,JJ))*ZCONC_TMP(JI,JJ)+PTOWN(JI,JJ)*XCONC_URBAN
      ZRAY(JI,JJ,JK)   = 0.5*((1.-PSEA(JI,JJ))*GAMMA(XNUC+1.0/XALPHAC)/(GAMMA(XNUC)) + &
              PSEA(JI,JJ)*GAMMA(XNUC2+1.0/XALPHAC2)/(GAMMA(XNUC2)))

    END DO
!$acc end kernels
#endif
  ELSE
!$acc kernels
        ZCONC3D(:,:,:) = XCONC_LAND
        ZRAY(:,:,:)  = 0.5*(GAMMA(XNUC+1.0/XALPHAC)/(GAMMA(XNUC)))
!$acc end kernels
   END IF
!$acc kernels
  ZRAY(:,:,:)      = MAX(1.,ZRAY(:,:,:))
  ZLBC(:,:,:)      = MAX(MIN(XLBC(1),XLBC(2)),ZLBC(:,:,:))
!$acc end kernels
ENDIF
!
!*       2.    compute the fluxes
!
!  optimization by looking for locations where
!  the precipitating fields are larger than a minimal value only !!!
!  For optimization we consider each variable separately
!
! External tendecies
IF (OSEDIC) THEN
!$acc kernels
  ZPRCS(:,:,:) = PRCS(:,:,:)-PRCT(:,:,:)*ZINVTSTEP
!$acc end kernels
ENDIF
!$acc kernels
ZPRRS(:,:,:) = PRRS(:,:,:)-PRRT(:,:,:)*ZINVTSTEP
ZPRIS(:,:,:) = PRIS(:,:,:)-PRIT(:,:,:)*ZINVTSTEP
ZPRSS(:,:,:) = PRSS(:,:,:)-PRST(:,:,:)*ZINVTSTEP
ZPRGS(:,:,:) = PRGS(:,:,:)-PRGT(:,:,:)*ZINVTSTEP
IF ( KRR == 7 ) THEN
  ZPRHS(:,:,:) = PRHS(:,:,:)-PRHT(:,:,:)*ZINVTSTEP
END IF
!
! mr values inside the time-splitting loop
ZRCT(:,:,:) = PRCT(:,:,:)
ZRRT(:,:,:) = PRRT(:,:,:)
ZRIT(:,:,:) = PRIT(:,:,:)
ZRST(:,:,:) = PRST(:,:,:)
ZRGT(:,:,:) = PRGT(:,:,:)
IF (KRR==7) THEN
  ZRHT(:,:,:) = PRHT(:,:,:)
END IF
!
ZW(:,:,KKTB:KKTE) =1./(PRHODREF(:,:,KKTB:KKTE)* PDZZ(:,:,KKTB:KKTE))
!$acc end kernels
!
!
!*       2.1   for cloud
!
IF (OSEDIC) THEN
    CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                          &XSPLIT_MAXCFL, &
                          &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                          &2, &
                          &ZRCT, PRCS, PINPRC, ZPRCS, &
                          &ZRAY, ZLBC, ZFSEDC, ZCONC3D, PFPR=PFPR)
ENDIF
!
!
!*       2.1bis  DROPLET DEPOSITION AT THE 1ST LEVEL ABOVE GROUND
!
!$acc kernels
IF (ODEPOSC) THEN
  PINDEP (:,:) = 0.
!$acc loop collapse(2) independent
  DO JJ=KJB,KJE
    DO JI=KIB,KIE
      IF (PRCS(JI,JJ,KKB)>0.) THEN
        PRCS(JI,JJ,KKB) = PRCS(JI,JJ,KKB) - PVDEPOSC * PRCT(JI,JJ,KKB) / PDZZ(JI,JJ,KKB)
        PINPRC(JI,JJ) = PINPRC(JI,JJ) + PVDEPOSC * PRCT(JI,JJ,KKB) * PRHODREF(JI,JJ,KKB) /XRHOLW
        PINDEP(JI,JJ) = PVDEPOSC * PRCT(JI,JJ,KKB) * PRHODREF(JI,JJ,KKB) /XRHOLW
      END IF
    END DO
  END DO
END IF
!$acc end kernels
!
!*       2.2   for rain
!
  CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                          &XSPLIT_MAXCFL, &
                          &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                          &3, &
                          &ZRRT, PRRS, PINPRR, ZPRRS, &
                          PFPR=PFPR)
!
!*       2.3   for pristine ice
!
  CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                          &XSPLIT_MAXCFL, &
                          &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                          &4, &
                          &ZRIT, PRIS, PINPRI, ZPRIS, &
                          PFPR=PFPR)
!
!*       2.4   for aggregates/snow
!
  CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                        &XSPLIT_MAXCFL, &
                        &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                        &5, &
                        &ZRST, PRSS, PINPRS, ZPRSS, &
                        PFPR=PFPR)
!
!*       2.5   for graupeln
!
  CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                        &XSPLIT_MAXCFL, &
                        &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                        &6, &
                        &ZRGT, PRGS, PINPRG, ZPRGS, &
                        PFPR=PFPR)
!
!*       2.6   for hail
!
IF (KRR==7) THEN
    CALL INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR, &
                          &XSPLIT_MAXCFL, &
                          &PRHODREF, ZW, PDZZ, PPABST, PTHT, PTSTEP, &
                          &7, &
                          &ZRHT, PRHS, PINPRH, ZPRHS, &
                          PFPR=PFPR)
ENDIF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"ICE4_SEDIMENTATION_SPLIT end:PRCS")
  CALL MPPDB_CHECK(PRRS,"ICE4_SEDIMENTATION_SPLIT end:PRRS")
  CALL MPPDB_CHECK(PRIS,"ICE4_SEDIMENTATION_SPLIT end:PRIS")
  CALL MPPDB_CHECK(PRSS,"ICE4_SEDIMENTATION_SPLIT end:PRSS")
  CALL MPPDB_CHECK(PRGS,"ICE4_SEDIMENTATION_SPLIT end:PRGS")
  IF (PRESENT(PRHS)) CALL MPPDB_CHECK(PRHS,"ICE4_SEDIMENTATION_SPLIT end:PRHS")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PINPRC,"ICE4_SEDIMENTATION_SPLIT end:PINPRC")
  CALL MPPDB_CHECK(PINDEP,"ICE4_SEDIMENTATION_SPLIT end:PINDEP")
  CALL MPPDB_CHECK(PINPRR,"ICE4_SEDIMENTATION_SPLIT end:PINPRR")
  CALL MPPDB_CHECK(PINPRI,"ICE4_SEDIMENTATION_SPLIT end:PINPRI")
  CALL MPPDB_CHECK(PINPRS,"ICE4_SEDIMENTATION_SPLIT end:PINPRS")
  CALL MPPDB_CHECK(PINPRG,"ICE4_SEDIMENTATION_SPLIT end:PINPRG")
  IF (PRESENT(PINPRH)) CALL MPPDB_CHECK(PINPRH,"ICE4_SEDIMENTATION_SPLIT end:PINPRH")
  IF (PRESENT(PFPR))   CALL MPPDB_CHECK(PFPR,  "ICE4_SEDIMENTATION_SPLIT end:PFPR")
END IF

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data
!
CONTAINS
!
!
!-------------------------------------------------------------------------------
!
!
SUBROUTINE INTERNAL_SEDIM_SPLI(KIB,KIE,KIT,KJB,KJE,KJT,KKB,KKTB,KKTE,KKT,KKL,KRR, &
                              &PMAXCFL,PRHODREF,POORHODZ,PDZZ,PPABST,PTHT,PTSTEP, &
                              &KSPE,PRXT,PRXS,PINPRX,PPRXS,                       &
                              &PRAY,PLBC,PFSEDC,PCONC3D,PFPR)
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XCPD,XP00,XRD
USE MODD_RAIN_ICE_DESCR, ONLY: XCC,XCEXVT,XDC,XLBEXC,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: XEXCSEDI,XEXSEDG,XEXSEDH,XEXSEDR,XEXSEDS,XFSEDG,XFSEDH,XFSEDI,XFSEDR,XFSEDS

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE

#endif
use mode_mppdb
use mode_msg

#if defined(MNH_BITREP) || defined(MNH_BITREP_OMP)
USE MODI_BITREP
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER, INTENT(IN) :: KIB,KIE,KIT, KJB,KJE,KJT, KKB, KKTB, KKTE, KKT, KKL, KRR
REAL,                         INTENT(IN)              :: PMAXCFL ! maximum CFL allowed
REAL, DIMENSION(:,:,:),       INTENT(IN)              :: PRHODREF ! Reference density
REAL, DIMENSION(:,:,KKTB : ), INTENT(IN)              :: POORHODZ ! One Over (Rhodref times delta Z)
REAL, DIMENSION(:,:,:),       INTENT(IN)              :: PDZZ ! layer thikness (m)
REAL, DIMENSION(:,:,:),       INTENT(IN)              :: PPABST
REAL, DIMENSION(:,:,:),       INTENT(IN)              :: PTHT
REAL,                         INTENT(IN)              :: PTSTEP  ! total timestep
INTEGER,                      INTENT(IN)              :: KSPE ! 1 for rc, 2 for rr...
REAL, DIMENSION(:,:,:),       INTENT(INOUT)           :: PRXT ! mr of specy X
REAL, DIMENSION(:,:,:),       INTENT(INOUT)           :: PRXS !Tendency of the specy KSPE
REAL, DIMENSION(:,:),         INTENT(OUT)             :: PINPRX ! instant precip
REAL, DIMENSION(:,:,:),       INTENT(IN)              :: PPRXS ! external tendencie
REAL, DIMENSION(:,:,:),       INTENT(IN), OPTIONAL    :: PRAY, PLBC, PFSEDC, PCONC3D
REAL, DIMENSION(:,:,:,:),     INTENT(INOUT), OPTIONAL :: PFPR    ! upper-air precipitation fluxes
!
!*       0.2  declaration of local variables
!
character(len=10)                              :: yspe ! String for error message
INTEGER                                        :: IDX, ISEDIM
INTEGER                                        :: JI, JJ, JK, JL
INTEGER, DIMENSION(:),     POINTER, CONTIGUOUS :: I1, I2, I3 ! Used to replace the COUNT
LOGICAL                                        :: GPRESENT_PFPR
REAL                                           :: ZINVTSTEP
REAL                                           :: ZZWLBDC, ZRAY, ZZT, ZZWLBDA, ZZCC
REAL                                           :: ZFSED, ZEXSED
REAL,    DIMENSION(:,:),   POINTER, CONTIGUOUS :: ZMRCHANGE
REAL,    DIMENSION(:,:),   POINTER, CONTIGUOUS :: ZMAX_TSTEP ! Maximum CFL in column
REAL,    DIMENSION(:),     POINTER, CONTIGUOUS :: ZRSMIN
REAL,    DIMENSION(:,:),   POINTER, CONTIGUOUS :: ZREMAINT                   ! Remaining time until the timestep end
REAL,    DIMENSION(:,:,:), POINTER, CONTIGUOUS :: ZWSED ! Sedimentation fluxes

!$acc data present(PRHODREF,POORHODZ,PDZZ,PPABST,PTHT,PRXT,PRXS,PINPRX,PPRXS, &
!$acc&             PRAY,PLBC,PFSEDC,PCONC3D,PFPR) &
!$acc&     copyin( XRTMIN )

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PRHODREF,"INTERNAL_SEDIM_SPLI beg:PRHODREF")
  CALL MPPDB_CHECK(POORHODZ,"INTERNAL_SEDIM_SPLI beg:POORHODZ")
  CALL MPPDB_CHECK(PDZZ,"INTERNAL_SEDIM_SPLI beg:PDZZ")
  CALL MPPDB_CHECK(PPABST,"INTERNAL_SEDIM_SPLI beg:PPABST")
  CALL MPPDB_CHECK(PTHT,"INTERNAL_SEDIM_SPLI beg:PTHT")
  CALL MPPDB_CHECK(PPRXS,"INTERNAL_SEDIM_SPLI beg:PPRXS")
  IF (PRESENT(PRAY))    CALL MPPDB_CHECK(PRAY,"INTERNAL_SEDIM_SPLI beg:PRAY")
  IF (PRESENT(PLBC))    CALL MPPDB_CHECK(PLBC,"INTERNAL_SEDIM_SPLI beg:PLBC")
  IF (PRESENT(PFSEDC))  CALL MPPDB_CHECK(PFSEDC,"INTERNAL_SEDIM_SPLI beg:PFSEDC")
  IF (PRESENT(PCONC3D)) CALL MPPDB_CHECK(PCONC3D,"INTERNAL_SEDIM_SPLI beg:PCONC3D")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRXT,"INTERNAL_SEDIM_SPLI beg:PRXT")
  CALL MPPDB_CHECK(PRXS,"INTERNAL_SEDIM_SPLI beg:PRXS")
  IF (PRESENT(PFPR))    CALL MPPDB_CHECK(PFPR,"INTERNAL_SEDIM_SPLI beg:PFPR")
END IF

#ifndef MNH_OPENACC
allocate( i1(kit * kjt * kkt ) )
allocate( i2(kit * kjt * kkt ) )
allocate( i3(kit * kjt * kkt ) )

allocate( zmrchange (kit, kjt ) )
allocate( zmax_tstep(kit, kjt ) )

allocate( zrsmin(size( xrtmin ) ) )

allocate( zremaint(size( prhodref, 1 ), size( prhodref, 2 ) ) )

allocate( zwsed(size( prhodref, 1 ), size( prhodref, 2 ), 0 : size( prhodref, 3 ) + 1 ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( i1, kit * kjt * kkt )
CALL MNH_MEM_GET( i2, kit * kjt * kkt )
CALL MNH_MEM_GET( i3, kit * kjt * kkt )

CALL MNH_MEM_GET( zmrchange,  kit, kjt )
CALL MNH_MEM_GET( zmax_tstep, kit, kjt )

CALL MNH_MEM_GET( zrsmin, Size( xrtmin ) )

CALL MNH_MEM_GET( zremaint, Size( prhodref, 1 ), Size( prhodref, 2 ) )

CALL MNH_MEM_GET( zwsed, 1, Size( prhodref, 1 ), 1, Size( prhodref, 2 ), 0, Size( prhodref, 3 ) + 1 )
#endif

!$acc data present( i1, i2, i3, zmrchange, zmax_tstep, zrsmin, zremaint, zwsed )

!
!-------------------------------------------------------------------------------
IF (KSPE<2 .OR. KSPE>7) CALL PRINT_MSG(NVERB_FATAL,'GEN','INTERNAL_SEDIM_SPLIT','invalid species (KSPE variable)')
!
IF (PRESENT(PFPR)) THEN
  GPRESENT_PFPR = .TRUE.
ELSE
  GPRESENT_PFPR = .FALSE.
END IF
!
!$acc kernels
PINPRX(:,:) = 0.
ZINVTSTEP=1./PTSTEP
ZRSMIN(:) = XRTMIN(:) * ZINVTSTEP
ZREMAINT(:,:) = PTSTEP
!$acc end kernels
!$acc update self(ZREMAINT)
!
DO WHILE (ANY(ZREMAINT>0.))
  ISEDIM = 0
!$acc parallel loop private(idx) independent
  DO JK = KKTB,KKTE
!$acc loop independent collapse(2)
    DO JJ = KJB,KJE
      DO JI = KIB,KIE
        IF( (PRXT (JI,JJ,JK)>XRTMIN(KSPE) .OR.    &
             PPRXS(JI,JJ,JK)>ZRSMIN(KSPE)) .AND. &
             ZREMAINT(JI,JJ)>0. ) THEN
!$acc atomic capture
          ISEDIM = ISEDIM + 1
          IDX = ISEDIM
!$acc end atomic
          I1(IDX) = JI
          I2(IDX) = JJ
          I3(IDX) = JK
        END IF
      END DO
    END DO
  END DO
!
  !
  !
  !*       1. Parameters for cloud sedimentation
  !
  !
  !*       2.    compute the fluxes
  !
  !
!acc kernels
  IF(KSPE==2) THEN
!$acc kernels
    !******* for cloud
    ZWSED(:,:,:) = 0.
!$acc loop independent private(JI,JJ,JK,JL,ZZWLBDC,ZRAY,ZZT,ZZWLBDA,ZZCC)
    DO JL=1, ISEDIM
      JI=I1(JL)
      JJ=I2(JL)
      JK=I3(JL)
      IF(PRXT(JI,JJ,JK)>XRTMIN(KSPE)) THEN
        ZZWLBDC = PLBC(JI,JJ,JK) * PCONC3D(JI,JJ,JK) / &
                  (PRHODREF(JI,JJ,JK) * PRXT(JI,JJ,JK))
#if !defined(MNH_BITREP) && !defined(MNH_BITREP_OMP)
        ZZWLBDC = ZZWLBDC**XLBEXC
        ZRAY = PRAY(JI,JJ,JK) / ZZWLBDC
        ZZT = PTHT(JI,JJ,JK) * (PPABST(JI,JJ,JK)/XP00)**(XRD/XCPD)
        ZZWLBDA = 6.6E-8*(101325./PPABST(JI,JJ,JK))*(ZZT/293.15)
        ZZCC = XCC*(1.+1.26*ZZWLBDA/ZRAY)
        ZWSED(JI, JJ, JK) = PRHODREF(JI,JJ,JK)**(-XCEXVT +1 ) *   &
                  ZZWLBDC**(-XDC)*ZZCC*PFSEDC(JI,JJ,JK) * PRXT(JI,JJ,JK)
#else
        ZZWLBDC = BR_POW(ZZWLBDC,XLBEXC)
        ZRAY = PRAY(JI,JJ,JK) / ZZWLBDC
        ZZT = PTHT(JI,JJ,JK) * BR_POW(PPABST(JI,JJ,JK)/XP00,XRD/XCPD)
        ZZWLBDA = 6.6E-8*(101325./PPABST(JI,JJ,JK))*(ZZT/293.15)
        ZZCC = XCC*(1.+1.26*ZZWLBDA/ZRAY)
        ZWSED(JI, JJ, JK) = BR_POW(PRHODREF(JI,JJ,JK),-XCEXVT+1) *   &
                  BR_POW(ZZWLBDC,-XDC)*ZZCC*PFSEDC(JI,JJ,JK) * PRXT(JI,JJ,JK)
#endif
      ENDIF
    ENDDO
!$acc end kernels
  ELSEIF(KSPE==4) THEN
!$acc kernels
    ! ******* for pristine ice
    ZWSED(:,:,:) = 0.
!$acc loop independent private(JI,JJ,JK,JL)
    DO JL=1, ISEDIM
      JI=I1(JL)
      JJ=I2(JL)
      JK=I3(JL)
      IF(PRXT(JI, JJ, JK) .GT. MAX(XRTMIN(4), 1.0E-7)) THEN
#if !defined(MNH_BITREP) && !defined(MNH_BITREP_OMP)
        ZWSED(JI, JJ, JK) =  XFSEDI * PRXT(JI, JJ, JK) *  &
                            & PRHODREF(JI,JJ,JK)**(1.-XCEXVT) * & !    McF&H
                            & MAX( 0.05E6,-0.15319E6-0.021454E6* &
                            &      ALOG(PRHODREF(JI,JJ,JK)*PRXT(JI,JJ,JK)) )**XEXCSEDI
#else
        ZWSED(JI, JJ, JK) =  XFSEDI * PRXT(JI, JJ, JK) *  &
                            & BR_POW(PRHODREF(JI,JJ,JK),1.-XCEXVT) * & !    McF&H
                            & BR_POW(MAX( 0.05E6,-0.15319E6-0.021454E6* &
                            &      BR_LOG(PRHODREF(JI,JJ,JK)*PRXT(JI,JJ,JK))),XEXCSEDI)
#endif
      ENDIF
    ENDDO
!$acc end kernels
  ELSE
!acc kernels
    ! ******* for other species
    SELECT CASE(KSPE)
      CASE(3)
!$acc kernels
        ZFSED=XFSEDR
        ZEXSED=XEXSEDR
!$acc end kernels
      CASE(5)
!$acc kernels
        ZFSED=XFSEDS
        ZEXSED=XEXSEDS
!$acc end kernels
      CASE(6)
!$acc kernels
        ZFSED=XFSEDG
        ZEXSED=XEXSEDG
!$acc end kernels
      CASE(7)
!$acc kernels
        ZFSED=XFSEDH
        ZEXSED=XEXSEDH
!$acc end kernels
      CASE DEFAULT
        write( yspe, '( I10 )' ) kspe
        call Print_msg( NVERB_FATAL, 'GEN', 'ICE4_SEDIMENTATION_SPLIT', 'no sedimentation parameter for KSPE='//trim(yspe) )
    END SELECT
    !
!$acc kernels
    ZWSED(:,:,:) = 0.
!$acc loop independent private(JI,JJ,JK,JL)
    DO JL=1, ISEDIM
      JI=I1(JL)
      JJ=I2(JL)
      JK=I3(JL)
      IF(PRXT(JI,JJ,JK)>XRTMIN(KSPE)) THEN
#if !defined(MNH_BITREP) && !defined(MNH_BITREP_OMP)
        ZWSED(JI, JJ, JK) = ZFSED  * PRXT(JI, JJ, JK)**ZEXSED            &
                                   * PRHODREF(JI, JJ, JK)**(ZEXSED-XCEXVT)
#else
        ZWSED(JI, JJ, JK) = ZFSED  * BR_POW(PRXT(JI, JJ, JK),ZEXSED)         &
                                   * BR_POW(PRHODREF(JI, JJ, JK),ZEXSED-XCEXVT)
#endif
      ENDIF
    ENDDO
!$acc end kernels
  ENDIF
!$acc kernels
  ZMAX_TSTEP(:,:) = ZREMAINT(:,:)
!acc loop independent private(JI,JJ,JK,JL)
  DO CONCURRENT( JL = 1 : ISEDIM )
    JI=I1(JL)
    JJ=I2(JL)
    JK=I3(JL)
    IF(PRXT(JI,JJ,JK)>XRTMIN(KSPE) .AND. ZWSED(JI, JJ, JK)>1.E-20) THEN
!acc atomic update
      ZMAX_TSTEP(JI, JJ) = MIN(ZMAX_TSTEP(JI, JJ), PMAXCFL * PRHODREF(JI, JJ, JK) * &
                           PRXT(JI, JJ, JK) * PDZZ(JI, JJ, JK) / ZWSED(JI, JJ, JK))
!acc end atomic
    ENDIF
  ENDDO
  ZREMAINT(:,:) = ZREMAINT(:,:) - ZMAX_TSTEP(:,:)
!$acc end kernels
!$acc kernels
  DO JK = KKTB , KKTE
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(2)
#endif
    DO CONCURRENT( JI = 1 : KIT, JJ = 1 : KJT )
      ZMRCHANGE(JI,JJ) = ZMAX_TSTEP(JI,JJ) * POORHODZ(JI,JJ,JK)*(ZWSED(JI,JJ,JK+KKL)-ZWSED(JI,JJ,JK))
      PRXT(JI,JJ,JK) = PRXT(JI,JJ,JK) + ZMRCHANGE(JI,JJ) + PPRXS(JI,JJ,JK) * ZMAX_TSTEP(JI,JJ)
      PRXS(JI,JJ,JK) = PRXS(JI,JJ,JK) + ZMRCHANGE(JI,JJ) * ZINVTSTEP
    END DO
  ENDDO
!$acc end kernels
!$acc kernels
  PINPRX(:,:) = PINPRX(:,:) + ZWSED(:,:,KKB) / XRHOLW * (ZMAX_TSTEP(:,:) * ZINVTSTEP)
  IF (GPRESENT_PFPR) THEN
!$acc loop independent
    DO JK = KKTB , KKTE
      PFPR(:,:,JK,KSPE) = PFPR(:,:,JK,KSPE) + ZWSED(:,:,JK) * (ZMAX_TSTEP(:,:) * ZINVTSTEP)
    ENDDO
  ENDIF
!$acc end kernels
!$acc update self(ZREMAINT)
!
END DO
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRXT,"INTERNAL_SEDIM_SPLI end:PRXT")
  CALL MPPDB_CHECK(PRXS,"INTERNAL_SEDIM_SPLI end:PRXS")
  IF (PRESENT(PFPR)) CALL MPPDB_CHECK(PFPR,"INTERNAL_SEDIM_SPLI beg:PFPR")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PINPRX,"INTERNAL_SEDIM_SPLI end:PINPRX")
END IF

!$acc end data

#ifndef MNH_OPENACC
deallocate( zwsed )
deallocate( zremaint )
deallocate( zrsmin )
deallocate( zmax_tstep )
deallocate( zmrchange )
deallocate( i3 )
deallocate( i2 )
deallocate( i1 )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data

END SUBROUTINE INTERNAL_SEDIM_SPLI
!
END SUBROUTINE ICE4_SEDIMENTATION_SPLIT
