!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #################### 
     MODULE MODI_TURB_VER_THERMO_CORR
!    ####################
!
INTERFACE 
!
      SUBROUTINE TURB_VER_THERMO_CORR(KKA,KKU,KKL,KRR,KRRL,KRRI,    &
                      OTURB_FLX,HTURBDIM,HTOM,                      &
                      PIMPL,PEXPL,                                  &
                      TPFILE,                                       &
                      PDZZ,                                         &
                      PWM,PTHLM,PRM,                                &
                      PTKEM,PLM,PLEPS,                              &
                      PATHETA,PAMOIST,                              &
                      PBETA, PSQRT_TKE, PDTH_DZ, PDR_DZ, PRED2TH3,  &
                      PRED2R3, PRED2THR3, PBLL_O_E, PETHETA,        &
                      PEMOIST, PREDTH1, PREDR1, PPHI3, PPSI3, PD,   &
                      PFWTH,PFWR,PFTH2,PFR2,PFTHR,                  &
                      PTHLP,PRP,PSIGS                          )
!
USE MODD_IO, ONLY: TFILEDATA
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=AR 
INTEGER,                INTENT(IN)   :: KRR           ! number of moist var.
INTEGER,                INTENT(IN)   :: KRRL          ! number of liquid water var.
INTEGER,                INTENT(IN)   :: KRRI          ! number of ice water var.
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
CHARACTER(len=4),       INTENT(IN)   ::  HTOM         ! type of Third Order Moment
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDZZ         ! Metric coefficients
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PWM 
! Vertical wind
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTHLM 
! potential temperature at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM          ! Mixing ratios 
                                                      ! at t-Delta t
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLEPS        ! dissipative length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PATHETA      ! coefficients between 
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PAMOIST      ! s and Thetal and Rnp
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBETA        ! buoyancy coefficient
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSQRT_TKE    ! sqrt(e)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDTH_DZ      ! d(th)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDR_DZ       ! d(rt)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2TH3     ! 3D Redeslperger number R*2_th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2R3      ! 3D Redeslperger number R*2_r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2THR3    ! 3D Redeslperger number R*2_thr
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBLL_O_E     ! beta * Lk * Leps / tke
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PETHETA      ! Coefficient for theta in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PEMOIST      ! Coefficient for r in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDTH1      ! 1D Redelsperger number for Th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDR1       ! 1D Redelsperger number for r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPHI3        ! Prandtl number for temperature
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPSI3        ! Prandtl number for vapor
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PD           ! Denominator in Prandtl numbers
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWTH        ! d(w'2th' )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWR         ! d(w'2r'  )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTH2        ! d(w'th'2 )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFR2         ! d(w'r'2  )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTHR        ! d(w'th'r')/dz (at mass point)
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHLP      ! guess of thl at t+ deltat
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRP        ! guess of r at t+ deltat
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)  ::  PSIGS     ! Vert. part of Sigma_s at t
!
END SUBROUTINE TURB_VER_THERMO_CORR
!
END INTERFACE
!
END MODULE MODI_TURB_VER_THERMO_CORR
!
!
!     ###############################################################
      SUBROUTINE TURB_VER_THERMO_CORR(KKA,KKU,KKL,KRR, KRRL, KRRI,  &
                      OTURB_FLX,HTURBDIM,HTOM,                      &
                      PIMPL,PEXPL,                                  &
                      TPFILE,                                       &
                      PDZZ,                                         &
                      PWM,PTHLM,PRM,                                &
                      PTKEM,PLM,PLEPS,                              &
                      PATHETA,PAMOIST,                              &
                      PBETA, PSQRT_TKE, PDTH_DZ, PDR_DZ, PRED2TH3,  &
                      PRED2R3, PRED2THR3, PBLL_O_E, PETHETA,        &
                      PEMOIST, PREDTH1, PREDR1, PPHI3, PPSI3, PD,   &
                      PFWTH,PFWR,PFTH2,PFR2,PFTHR,                  &
                      PTHLP,PRP,PSIGS                          )
!     ###############################################################
!
!
!!****  *TURB_VER_THERMO_FLUX* -compute the source terms due to the vertical turbulent
!!       fluxes.
!!
!!    PURPOSE
!!    -------
!       The purpose of this routine is to compute the vertical turbulent
!     fluxes of the evolutive variables and give back the source 
!     terms to the main program.	In the case of large horizontal meshes,
!     the divergence of these vertical turbulent fluxes represent the whole
!     effect of the turbulence but when the three-dimensionnal version of
!     the turbulence scheme is activated (CTURBDIM="3DIM"), these divergences
!     are completed in the next routine TURB_HOR. 
!		  An arbitrary degree of implicitness has been implemented for the 
!     temporal treatment of these diffusion terms.
!       The vertical boundary conditions are as follows:
!           *  at the bottom, the surface fluxes are prescribed at the same
!              as the other turbulent fluxes
!           *  at the top, the turbulent fluxes are set to 0.
!       It should be noted that the condensation has been implicitely included
!     in this turbulence scheme by using conservative variables and computing
!     the subgrid variance of a statistical variable s indicating the presence 
!     or not of condensation in a given mesh. 
!
!!**  METHOD
!!    ------
!!      1D type calculations are made;
!!      The vertical turbulent fluxes are computed in an off-centered
!!      implicit scheme (a Crank-Nicholson type with coefficients different
!!      than 0.5), which allows to vary the degree of implicitness of the
!!      formulation.
!!      	 The different prognostic variables are treated one by one. 
!!      The contributions of each turbulent fluxes are cumulated into the 
!!      tendency  PRvarS, and into the dynamic and thermal production of 
!!      TKE if necessary.
!!        
!!			 In section 2 and 3, the thermodynamical fields are considered.
!!      Only the turbulent fluxes of the conservative variables
!!      (Thetal and Rnp stored in PRx(:,:,:,1))  are computed. 
!!       Note that the turbulent fluxes at the vertical 
!!      boundaries are given either by the soil scheme for the surface one
!!      ( at the same instant as the others fluxes) and equal to 0 at the 
!!      top of the model. The thermal production is computed by vertically 
!!      averaging the turbulent flux and multiply this flux at the mass point by
!!      a function ETHETA or EMOIST, which preform the transformation from the
!!      conservative variables to the virtual potential temperature. 
!!     
!! 	    In section 4, the variance of the statistical variable
!!      s indicating presence or not of condensation, is determined in function 
!!      of the turbulent moments of the conservative variables and its
!!      squarred root is stored in PSIGS. This information will be completed in 
!!      the horizontal turbulence if the turbulence dimensionality is not 
!!      equal to "1DIM".
!!
!!			 In section 5, the x component of the stress tensor is computed.
!!      The surface flux <u'w'> is computed from the value of the surface
!!      fluxes computed in axes linked to the orography ( i", j" , k"):
!!        i" is parallel to the surface and in the direction of the maximum
!!           slope
!!        j" is also parallel to the surface and in the normal direction of
!!           the maximum slope
!!        k" is the normal to the surface
!!      In order to prevent numerical instability, the implicit scheme has 
!!      been extended to the surface flux regarding to its dependence in 
!!      function of U. The dependence in function of the other components 
!!      introduced by the different rotations is only explicit.
!!      The turbulent fluxes are used to compute the dynamic production of 
!!      TKE. For the last TKE level ( located at PDZZ(:,:,IKB)/2 from the
!!      ground), an harmonic extrapolation from the dynamic production at 
!!      PDZZ(:,:,IKB) is used to avoid an evaluation of the gradient of U
!!      in the surface layer.
!!
!!         In section 6, the same steps are repeated but for the y direction
!!		  and in section 7, a diagnostic computation of the W variance is 
!!      performed.
!!
!!         In section 8, the turbulent fluxes for the scalar variables are 
!!      computed by the same way as the conservative thermodynamical variables
!!
!!            
!!    EXTERNAL
!!    --------
!!      GX_U_M, GY_V_M, GZ_W_M :  cartesian gradient operators 
!!      GX_U_UW,GY_V_VW	         (X,Y,Z) represent the direction of the gradient
!!                               _(M,U,...)_ represent the localization of the 
!!                               field to be derivated
!!                               _(M,UW,...) represent the localization of the 
!!                               field	derivated
!!                               
!!
!!      MXM,MXF,MYM,MYF,MZM,MZF
!!                             :  Shuman functions (mean operators)     
!!      DXF,DYF,DZF,DZM
!!                             :  Shuman functions (difference operators)     
!!                               
!!      SUBROUTINE TRIDIAG     : to compute the split implicit evolution
!!                               of a variable located at a mass point
!!
!!      SUBROUTINE TRIDIAG_WIND: to compute the split implicit evolution
!!                               of a variable located at a wind point
!!
!!      FUNCTIONs ETHETA and EMOIST  :  
!!            allows to compute:
!!            - the coefficients for the turbulent correlation between
!!            any variable and the virtual potential temperature, of its 
!!            correlations with the conservative potential temperature and 
!!            the humidity conservative variable:
!!            -------              -------              -------
!!            A' Thv'  =  ETHETA   A' Thl'  +  EMOIST   A' Rnp'  
!!
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      Module MODD_CST : contains physical constants
!!
!!           XG         : gravity constant
!!
!!      Module MODD_CTURB: contains the set of constants for
!!                        the turbulence scheme
!!
!!           XCMFS,XCMFB : cts for the momentum flux
!!           XCSHF       : ct for the sensible heat flux
!!           XCHF        : ct for the moisture flux
!!           XCTV,XCHV   : cts for the T and moisture variances
!!
!!      Module MODD_PARAMETERS
!!
!!           JPVEXT_TURB     : number of vertical external points
!!           JPHEXT     : number of horizontal external points
!!
!!
!!    REFERENCE
!!    ---------
!!      Book 1 of documentation (Chapter: Turbulence)
!!
!!    AUTHOR
!!    ------
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original       August   19, 1994
!!      Modifications: February 14, 1995 (J.Cuxart and J.Stein) 
!!                                  Doctorization and Optimization
!!      Modifications: March 21, 1995 (J.M. Carriere) 
!!                                  Introduction of cloud water
!!      Modifications: June  14, 1995 (J.Cuxart and J. Stein) 
!!                                 Phi3 and Psi3 at w-point + bug in the all
!!                                 or nothing condens. 
!!      Modifications: Sept  15, 1995 (J.Cuxart and J. Stein) 
!!                                 Change the DP computation at the ground
!!      Modifications: October 10, 1995 (J.Cuxart and J. Stein) 
!!                                 Psi for scal var and LES tools
!!      Modifications: November 10, 1995 (J. Stein)
!!                                 change the surface	relations 
!!      Modifications: February 20, 1995 (J. Stein) optimization
!!      Modifications: May 21, 1996 (J. Stein) 
!!                                  bug in the vertical flux of the V wind 
!!                                  component for explicit computation
!!      Modifications: May 21, 1996 (N. wood) 
!!                                  modify the computation of the vertical
!!                                   part or the surface tangential flux
!!      Modifications: May 21, 1996 (P. Jabouille)
!!                                  same modification in the Y direction
!!      
!!      Modifications: Sept 17, 1996 (J. Stein) change the moist case by using
!!                                  Pi instead of Piref + use Atheta and Amoist
!!
!!      Modifications: Nov  24, 1997 (V. Masson) removes the DO loops 
!!      Modifications: Mar  31, 1998 (V. Masson) splits the routine TURB_VER_THERMO_FLUX 
!!      Modifications: Oct  18, 2000 (V. Masson) LES computations
!!      Modifications: Dec  01, 2000 (V. Masson) conservation of energy from
!!                                               surface flux in 1DIM case
!!                                               when slopes are present
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     October 2009 (G. Tanguy) add ILENCH=LEN(YCOMMENT) after
!!                                              change of YCOMMENT
!!                     2012-02 (Y. Seity) add possibility to run with reversed 
!!                                              vertical levels
!!                     04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!!  Philippe Wautelet: 05/2016-04/2018: new data structures and calls for I/O
!!--------------------------------------------------------------------------
!       
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_CTURB
use modd_field,          only: tfielddata, TYPEREAL
USE MODD_IO,             ONLY: TFILEDATA
USE MODD_PARAMETERS
USE MODD_CONF
USE MODD_LES
!
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
USE MODI_GRADIENT_M
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_TRIDIAG 
USE MODI_LES_MEAN_SUBGRID
!
USE MODE_IO_FIELD_WRITE, only: IO_Field_write
use mode_mppdb
USE MODE_PRANDTL
!
USE MODI_SECOND_MNH
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif

IMPLICIT NONE
!
!*      0.1  declarations of arguments
!
!
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=ARO
INTEGER,                INTENT(IN)   :: KRR           ! number of moist var.
INTEGER,                INTENT(IN)   :: KRRL          ! number of liquid water var.
INTEGER,                INTENT(IN)   :: KRRI          ! number of ice water var.
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
CHARACTER(len=4),       INTENT(IN)   ::  HTOM         ! type of Third Order Moment
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDZZ         ! Metric coefficients
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PWM 
! Vertical wind
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTHLM 
! potential temperature at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM          ! Mixing ratios 
                                                      ! at t-Delta t
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLEPS        ! dissipative length   
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PATHETA      ! coefficients between 
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PAMOIST      ! s and Thetal and Rnp
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBETA        ! buoyancy coefficient
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PSQRT_TKE    ! sqrt(e)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDTH_DZ      ! d(th)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDR_DZ       ! d(rt)/dz
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2TH3     ! 3D Redeslperger number R*2_th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2R3      ! 3D Redeslperger number R*2_r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRED2THR3    ! 3D Redeslperger number R*2_thr
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PBLL_O_E     ! beta * Lk * Leps / tke
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PETHETA      ! Coefficient for theta in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PEMOIST      ! Coefficient for r in theta_v computation
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDTH1      ! 1D Redelsperger number for Th
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PREDR1       ! 1D Redelsperger number for r
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPHI3        ! Prandtl number for temperature
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PPSI3        ! Prandtl number for vapor
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PD           ! Denominator in Prandtl numbers
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWTH        ! d(w'2th' )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFWR         ! d(w'2r'  )/dz (at flux point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTH2        ! d(w'th'2 )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFR2         ! d(w'r'2  )/dz (at mass point)
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PFTHR        ! d(w'th'r')/dz (at mass point)
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHLP      ! guess of thl at t+ deltat
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRP        ! guess of r at t+ deltat
!
REAL, DIMENSION(:,:,:),   INTENT(OUT)  ::  PSIGS     ! Vert. part of Sigma_s at t
!
!*       0.2  declaration of local variables
!
!
REAL, DIMENSION(:,:,:), pointer , contiguous ::  &
       ZFLXZ,    & ! vertical flux of the treated variable
       ZKEFF,    & ! effectif diffusion coeff = LT * SQRT( TKE )
       ZF,       & ! Flux in dTh/dt =-dF/dz (evaluated at t-1)(or rt instead of Th)
       ZDFDDTDZ, & ! dF/d(dTh/dz)
       ZDFDDRDZ    ! dF/d(dr/dz)
INTEGER             :: IKB,IKE      ! I index values for the Beginning and End
                                    ! mass points of the domain in the 3 direct.
INTEGER             :: I1,I2        ! For ZCOEFF allocation
REAL, DIMENSION(:,:,:),POINTER , CONTIGUOUS  :: ZCOEFF
                                    ! coefficients for the uncentred gradient 
                                    ! computation near the ground
!
REAL :: ZTIME1, ZTIME2
!
LOGICAL :: GUSERV   ! flag to use water
LOGICAL :: GFTH2    ! flag to use w'th'2
LOGICAL :: GFWTH    ! flag to use w'2th'
LOGICAL :: GFR2     ! flag to use w'r'2
LOGICAL :: GFWR     ! flag to use w'2r'
LOGICAL :: GFTHR    ! flag to use w'th'r'
!
#ifdef MNH_OPENACC
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP1_DEVICE,ZTMP2_DEVICE,ZTMP3_DEVICE,ZTMP4_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP5_DEVICE,ZTMP6_DEVICE,ZTMP7_DEVICE,ZTMP8_DEVICE
#endif
TYPE(TFIELDDATA) :: TZFIELD
!
INTEGER  :: JIU,JJU,JKU
INTEGER  :: JI,JJ,JK
!----------------------------------------------------------------------------

!$acc data present( PDZZ,                                            &
!$acc &             PWM, PTHLM, PRM,                                 &
!$acc &             PTKEM, PLM, PLEPS,                               &
!$acc &             PATHETA, PAMOIST,                                &
!$acc &             PBETA,  PSQRT_TKE,  PDTH_DZ,  PDR_DZ,  PRED2TH3, &
!$acc &             PRED2R3,  PRED2THR3,  PBLL_O_E,  PETHETA,        &
!$acc &             PEMOIST,  PREDTH1,  PREDR1,  PPHI3,  PPSI3,  PD, &
!$acc &             PFWTH, PFWR, PFTH2, PFR2, PFTHR,                 &
!$acc &             PTHLP, PRP, PSIGS                                )

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( pdzz,      "Turb_ver_thermo_corr beg:pdzz"      )
  call Mppdb_check( pwm,       "Turb_ver_thermo_corr beg:pwm"       )
  call Mppdb_check( pthlm,     "Turb_ver_thermo_corr beg:pthlm"     )
  call Mppdb_check( prm,       "Turb_ver_thermo_corr beg:prm"       )
  call Mppdb_check( ptkem,     "Turb_ver_thermo_corr beg:ptkem"     )
  call Mppdb_check( plm,       "Turb_ver_thermo_corr beg:plm"       )
  call Mppdb_check( pleps,     "Turb_ver_thermo_corr beg:pleps"     )
  call Mppdb_check( patheta,   "Turb_ver_thermo_corr beg:patheta"   )
  call Mppdb_check( pamoist,   "Turb_ver_thermo_corr beg:pamoist"   )
  call Mppdb_check( pbeta,     "Turb_ver_thermo_corr beg:pbeta"     )
  call Mppdb_check( psqrt_tke, "Turb_ver_thermo_corr beg:psqrt_tke" )
  call Mppdb_check( pdth_dz,   "Turb_ver_thermo_corr beg:pdth_dz"   )
  call Mppdb_check( pdr_dz,    "Turb_ver_thermo_corr beg:pdr_dz"    )
  call Mppdb_check( pred2th3,  "Turb_ver_thermo_corr beg:pred2th3"  )
  call Mppdb_check( pred2r3,   "Turb_ver_thermo_corr beg:pred2r3"   )
  call Mppdb_check( pred2thr3, "Turb_ver_thermo_corr beg:pred2thr3" )
  call Mppdb_check( pbll_o_e,  "Turb_ver_thermo_corr beg:pbll_o_e"  )
  call Mppdb_check( petheta,   "Turb_ver_thermo_corr beg:petheta"   )
  call Mppdb_check( pemoist,   "Turb_ver_thermo_corr beg:pemoist"   )
  call Mppdb_check( predth1,   "Turb_ver_thermo_corr beg:predth1"   )
  call Mppdb_check( predr1,    "Turb_ver_thermo_corr beg:predr1"    )
  call Mppdb_check( pphi3,     "Turb_ver_thermo_corr beg:pphi3"     )
  call Mppdb_check( ppsi3,     "Turb_ver_thermo_corr beg:ppsi3"     )
  call Mppdb_check( pd,        "Turb_ver_thermo_corr beg:pd"        )
  call Mppdb_check( pfwth,     "Turb_ver_thermo_corr beg:pfwth"     )
  call Mppdb_check( pfwr,      "Turb_ver_thermo_corr beg:pfwr"      )
  call Mppdb_check( pfth2,     "Turb_ver_thermo_corr beg:pfth2"     )
  call Mppdb_check( pfr2,      "Turb_ver_thermo_corr beg:pfr2"      )
  call Mppdb_check( pfthr,     "Turb_ver_thermo_corr beg:pfthr"     )
  call Mppdb_check( pthlp,     "Turb_ver_thermo_corr beg:pthlp"     )
  call Mppdb_check( prp,       "Turb_ver_thermo_corr beg:prp"       )
end if
!
!*       1.   PRELIMINARIES
!             -------------
!
IKB=KKA+JPVEXT_TURB*KKL
IKE=KKU-JPVEXT_TURB*KKL
I1=MIN(KKA+JPVEXT_TURB*KKL,KKA+JPVEXT_TURB*KKL+2*KKL)
I2=MAX(KKA+JPVEXT_TURB*KKL,KKA+JPVEXT_TURB*KKL+2*KKL)

JIU =  size( pthlm, 1 )
JJU =  size( pthlm, 2 )
JKU =  size( pthlm, 3 )

#ifndef MNH_OPENACC
allocate( zflxz     (JIU,JJU,JKU) )
allocate( zkeff     (JIU,JJU,JKU) )
allocate( zf        (JIU,JJU,JKU) )
allocate( zdfddtdz  (JIU,JJU,JKU) )
allocate( zdfddrdz  (JIU,JJU,JKU) )

allocate( zcoeff(JIU,JJU, i1 : i2 ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( zflxz,    JIU, JJU, JKU )
CALL MNH_MEM_GET( zkeff,    JIU, JJU, JKU )
CALL MNH_MEM_GET( zf,       JIU, JJU, JKU )
CALL MNH_MEM_GET( zdfddtdz, JIU, JJU, JKU )
CALL MNH_MEM_GET( zdfddrdz, JIU, JJU, JKU )

CALL MNH_MEM_GET( zcoeff, 1, JIU, 1, JJU, i1, i2 )

CALL MNH_MEM_GET( ztmp1_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp2_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp3_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp4_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp5_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp6_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp7_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp8_device, JIU, JJU, JKU )
#endif

!$acc data present( zflxz, zkeff, zf, zdfddtdz, zdfddrdz, zcoeff,           &
!$acc &            ztmp1_device, ztmp2_device, ztmp3_device, ztmp4_device, &
!$acc &            ztmp5_device, ztmp6_device, ztmp7_device, ztmp8_device  )

GUSERV = (KRR/=0)
!
!  compute the coefficients for the uncentred gradient computation near the 
!  ground
!$acc kernels present_cr(ZCOEFF,ZKEFF,ZTMP1_DEVICE)
ZCOEFF(:,:,IKB+2*KKL)= - PDZZ(:,:,IKB+KKL) /      &
       ( (PDZZ(:,:,IKB+2*KKL)+PDZZ(:,:,IKB+KKL)) * PDZZ(:,:,IKB+2*KKL) )
ZCOEFF(:,:,IKB+KKL)=   (PDZZ(:,:,IKB+2*KKL)+PDZZ(:,:,IKB+KKL)) /      &
       ( PDZZ(:,:,IKB+KKL) * PDZZ(:,:,IKB+2*KKL) )
ZCOEFF(:,:,IKB)= - (PDZZ(:,:,IKB+2*KKL)+2.*PDZZ(:,:,IKB+KKL)) /      &
       ( (PDZZ(:,:,IKB+2*KKL)+PDZZ(:,:,IKB+KKL)) * PDZZ(:,:,IKB+KKL) )
!
#ifndef MNH_OPENACC
ZKEFF(:,:,:) = MZM( PLM(:,:,:) * SQRT(PTKEM(:,:,:)) )
#else
!dir$ concurrent
ZTMP1_DEVICE(:,:,:) = PLM(:,:,:) * SQRT(PTKEM(:,:,:))
!$acc end kernels
CALL MZM_DEVICE(ZTMP1_DEVICE(:,:,:), ZKEFF(:,:,:))
#endif
!
! Flags for 3rd order quantities
!
GFTH2 = .FALSE.
GFR2  = .FALSE.
GFTHR = .FALSE.
GFWTH = .FALSE.
GFWR  = .FALSE.
!
IF (HTOM/='NONE') THEN
  GFTH2 = ANY(PFTH2(:,:,:)/=0.)
  GFR2  = ANY(PFR2(:,:,:) /=0.) .AND. GUSERV
  GFTHR = ANY(PFTHR(:,:,:)/=0.) .AND. GUSERV
  GFWTH = ANY(PFWTH(:,:,:)/=0.)
  GFWR  = ANY(PFWR(:,:,:) /=0.) .AND. GUSERV
END IF
!----------------------------------------------------------------------------
!
!
!*       4.   TURBULENT CORRELATIONS : <THl THl>, <THl Rnp>, <Rnp Rnp>
!             --------------------------------------------------------
!
!
!*       4.2  <THl THl> 
!
! Compute the turbulent variance F and F' at time t-dt.
#ifndef MNH_OPENACC
#ifndef MNH_BITREP
  ZF      (:,:,:) = XCTV*PLM(:,:,:)*PLEPS(:,:,:)*MZF(PPHI3(:,:,:)*PDTH_DZ(:,:,:)**2)
#else
  ZF      (:,:,:) = XCTV*PLM(:,:,:)*PLEPS(:,:,:)*MZF(PPHI3(:,:,:)*BR_P2(PDTH_DZ(:,:,:)))
#endif
  ZDFDDTDZ(:,:,:) = 0.     ! this term, because of discretization, is treated separately
#else
  !$acc kernels present_cr(ZTMP1_DEVICE)
#ifndef MNH_BITREP
  !dir$ concurrent
  ZTMP1_DEVICE(:,:,:) = PPHI3(:,:,:)*PDTH_DZ(:,:,:)**2
#else
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP1_DEVICE(JI,JJ,JK) = PPHI3(JI,JJ,JK)*BR_P2(PDTH_DZ(JI,JJ,JK))
  END DO
#endif
  !$acc end kernels
  CALL MZF_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP2_DEVICE(:,:,:) )
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZF      (JI,JJ,JK) = XCTV*PLM(JI,JJ,JK)*PLEPS(JI,JJ,JK)*ZTMP2_DEVICE(JI,JJ,JK) 
  END DO
  !$acc end kernels
  !$acc kernels present_cr(ZDFDDTDZ)
  ZDFDDTDZ(:,:,:) = 0.     ! this term, because of discretization, is treated separately 
  !$acc end kernels
#endif
  !
  ! Effect of 3rd order terms in temperature flux (at mass point)
  !
  ! d(w'th'2)/dz
#ifndef MNH_OPENACC
  IF (GFTH2) THEN
    ZF(:,:,:)       = ZF(:,:,:)       + M3_TH2_WTH2(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
     & PSQRT_TKE(:,:,:)) * PFTH2(:,:,:)
    ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_TH2_WTH2_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
     & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:)) * PFTH2(:,:,:)
  END IF
#else
  IF (GFTH2) THEN
    CALL M3_TH2_WTH2( PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),ZTMP1_DEVICE(:,:,:))
    !$acc kernels present_cr(ZF)
    ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTH2(:,:,:)
    !$acc end kernels
    CALL D_M3_TH2_WTH2_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
     & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),ZTMP2_DEVICE(:,:,:))
    !$acc kernels present_cr(ZDFDDTDZ)
    ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFTH2(:,:,:)
    !$acc end kernels
  END IF
#endif
  !
  ! d(w'2th')/dz
#ifndef MNH_OPENACC
  IF (GFWTH) THEN
    ZF(:,:,:)       = ZF(:,:,:)       + M3_TH2_W2TH(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),PDTH_DZ(:,:,:),&
     & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:)) * MZF(PFWTH(:,:,:))
    ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_TH2_W2TH_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
     & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),GUSERV) * MZF(PFWTH(:,:,:))
  END IF
#else
  IF (GFWTH) THEN
    CALL MZF_DEVICE( PFWTH(:,:,:), ZTMP1_DEVICE(:,:,:) )
    CALL M3_TH2_W2TH( PREDTH1(:,:,:), PREDR1(:,:,:), PD(:,:,:), PDTH_DZ(:,:,:), PLM(:,:,:), &
                      PLEPS(:,:,:), PTKEM(:,:,:), ZTMP2_DEVICE(:,:,:) )
    !$acc kernels present_cr(ZF)
    ZF(:,:,:)       = ZF(:,:,:)       + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
    !$acc end kernels
    CALL D_M3_TH2_W2TH_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
     & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),GUSERV,ZTMP2_DEVICE(:,:,:))
    !$acc kernels present_cr(ZDFDDTDZ)
    ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
    !$acc end kernels
  END IF
#endif
  !
  IF (KRR/=0) THEN
    ! d(w'r'2)/dz
#ifndef MNH_OPENACC
    IF (GFR2) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_TH2_WR2(PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),&
       & PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFR2(:,:,:)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_TH2_WR2_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFR2(:,:,:)
    END IF
#else
    IF (GFR2) THEN
      CALL M3_TH2_WR2( PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),&
       & PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
      CALL D_M3_TH2_WR2_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'2r')/dz
#ifndef MNH_OPENACC
    IF (GFWR) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_TH2_W2R(PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),&
       & PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * MZF(PFWR(:,:,:))
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_TH2_W2R_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
       & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * MZF(PFWR(:,:,:))
    END IF
#else
    IF (GFWR) THEN
      CALL MZF_DEVICE( PFWR(:,:,:), ZTMP1_DEVICE(:,:,:) )
      CALL M3_TH2_W2R( PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),&
       & PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
      CALL D_M3_TH2_W2R_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),&
       & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP3_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP3_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'th'r')/dz
#ifndef MNH_OPENACC
    IF (GFTHR) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_TH2_WTHR(PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFTHR(:,:,:)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_TH2_WTHR_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFTHR(:,:,:)
    END IF
#else
    IF (GFTHR) THEN
      CALL M3_TH2_WTHR( PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTHR(:,:,:)
      !$acc end kernels
      CALL D_M3_TH2_WTHR_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFTHR(:,:,:)
      !$acc end kernels
    END IF
#endif

  END IF
  !
#ifndef MNH_OPENACC
  ZFLXZ(:,:,:)   = ZF(:,:,:)                                                              &
  !     + PIMPL * XCTV*PLM(:,:,:)*PLEPS(:,:,:)                                                   &
  !        *MZF(D_PHI3DTDZ2_O_DDTDZ(PPHI3(:,:,:),PREDTH1(:,:,:),PREDR1(:,:,:),PRED2TH3(:,:,:),&
  !                                             PRED2THR3(:,:,:),PDTH_DZ(:,:,:),HTURBDIM,GUSERV)   &
  !             *DZM(PTHLP(:,:,:) - PTHLM(:,:,:)) / PDZZ(:,:,:)                                        ) &
        + PIMPL * ZDFDDTDZ(:,:,:) * MZF(DZM(PTHLP(:,:,:) - PTHLM(:,:,:)) / PDZZ(:,:,:) )
#else
  !$acc kernels present_cr(ZTMP1_DEVICE)
  ZTMP1_DEVICE(:,:,:) = PTHLP(:,:,:) - PTHLM(:,:,:)
  !$acc end kernels
  CALL DZM_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP2_DEVICE(:,:,:) )
  !$acc kernels
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP3_DEVICE(JI,JJ,JK) = ZTMP2_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
  END DO
  !$acc end kernels
  CALL MZF_DEVICE( ZTMP3_DEVICE(:,:,:), ZTMP4_DEVICE(:,:,:) )
#ifdef MNH_COMPILER_CCE
  !PASVSCAL$acc kernels present(ZFLXZ)
  !$acc kernels 
#else
  !$acc kernels 
#endif
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(3)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZFLXZ(JI,JJ,JK)   = ZF(JI,JJ,JK)                                                              &
          + PIMPL * ZDFDDTDZ(JI,JJ,JK) * ZTMP4_DEVICE(JI,JJ,JK)
  END DO
!$acc end kernels
#endif
  !
  ! special case near the ground ( uncentred gradient )
#ifndef MNH_BITREP
  !$acc kernels  present_cr(ZFLXZ)
  ZFLXZ(:,:,IKB) = XCTV * PPHI3(:,:,IKB+KKL) * PLM(:,:,IKB)   &
     * PLEPS(:,:,IKB)                                         &
  *( PEXPL *                                                  &
     ( ZCOEFF(:,:,IKB+2*KKL)*PTHLM(:,:,IKB+2*KKL)             &
      +ZCOEFF(:,:,IKB+KKL  )*PTHLM(:,:,IKB+KKL  )             & 
      +ZCOEFF(:,:,IKB      )*PTHLM(:,:,IKB  )   )**2          &
    +PIMPL *                                                  &
     ( ZCOEFF(:,:,IKB+2*KKL)*PTHLP(:,:,IKB+2*KKL)             &
      +ZCOEFF(:,:,IKB+KKL  )*PTHLP(:,:,IKB+KKL  )             &
      +ZCOEFF(:,:,IKB      )*PTHLP(:,:,IKB  )   )**2          &
   ) 
  !$acc end kernels
#else
  !$acc kernels 
#ifdef MNH_COMPILER_NVHPC
  !$acc loop independent collapse(2)
#endif
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
  ZFLXZ(JI,JJ,IKB) = XCTV * PPHI3(JI,JJ,IKB+KKL) * PLM(JI,JJ,IKB)   &
     * PLEPS(JI,JJ,IKB)                                         &
  *( PEXPL *                                                  &
     BR_P2( ZCOEFF(JI,JJ,IKB+2*KKL)*PTHLM(JI,JJ,IKB+2*KKL)        &
      +ZCOEFF(JI,JJ,IKB+KKL  )*PTHLM(JI,JJ,IKB+KKL  )             & 
      +ZCOEFF(JI,JJ,IKB      )*PTHLM(JI,JJ,IKB  )   )             &
    +PIMPL *                                                  &
     BR_P2( ZCOEFF(JI,JJ,IKB+2*KKL)*PTHLP(JI,JJ,IKB+2*KKL)        &
      +ZCOEFF(JI,JJ,IKB+KKL  )*PTHLP(JI,JJ,IKB+KKL  )             &
      +ZCOEFF(JI,JJ,IKB      )*PTHLP(JI,JJ,IKB  )   )             &
      )
   END DO
  !$acc end kernels
#endif

  !$acc kernels present_cr(ZFLXZ,PSIGS)
  ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
  !
  ZFLXZ(:,:,:) = MAX(0., ZFLXZ(:,:,:))
  IF (KRRL > 0)  THEN
#ifndef MNH_BITREP
    PSIGS(:,:,:) = ZFLXZ(:,:,:) * PATHETA(:,:,:)**2 
#else
    PSIGS(:,:,:) = ZFLXZ(:,:,:) * BR_P2(PATHETA(:,:,:)) 
#endif
  END IF
!$acc end kernels
  !
  !
  ! stores <THl THl>  
  IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
    TZFIELD%CMNHNAME   = 'THL_VVAR'
    TZFIELD%CSTDNAME   = ''
    TZFIELD%CLONGNAME  = 'THL_VVAR'
    TZFIELD%CUNITS     = 'K2'
    TZFIELD%CDIR       = 'XY'
    TZFIELD%CCOMMENT   = 'X_Y_Z_THL_VVAR'
    TZFIELD%NGRID      = 1
    TZFIELD%NTYPE      = TYPEREAL
    TZFIELD%NDIMS      = 3
    TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLXZ)
    CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ(:,:,:))
  END IF
!
! and we store in LES configuration
!
  IF (LLES_CALL) THEN
    CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
    CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_Thl2 )
    CALL LES_MEAN_SUBGRID( MZF(PWM(:,:,:))*ZFLXZ(:,:,:), X_LES_RES_W_SBG_Thl2 )
    CALL LES_MEAN_SUBGRID( -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:), X_LES_SUBGRID_DISS_Thl2 )
    CALL LES_MEAN_SUBGRID( PETHETA(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_ThlThv )
    CALL LES_MEAN_SUBGRID( -XA3*PBETA(:,:,:)*PETHETA(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_ThlPz, .TRUE. )
#else
!$acc data copy(X_LES_SUBGRID_Thl2,X_LES_RES_W_SBG_Thl2,X_LES_SUBGRID_DISS_Thl2, &
!$acc &         X_LES_SUBGRID_ThlThv,X_LES_SUBGRID_ThlPz)
    CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_Thl2 )
    !
    CALL  MZF_DEVICE( PWM(:,:,:), ZTMP1_DEVICE(:,:,:) )
    !$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE(:,:,:), X_LES_RES_W_SBG_Thl2 )
    !
    !$acc kernels present_cr(ZTMP1_DEVICE)
    ZTMP1_DEVICE(:,:,:) = -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_DISS_Thl2 )
    !
    !$acc kernels present_cr(ZTMP1_DEVICE)
    ZTMP1_DEVICE(:,:,:) = PETHETA(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_ThlThv )
    !
    !$acc kernels present_cr(ZTMP1_DEVICE)
    ZTMP1_DEVICE(:,:,:) = -XA3*PBETA(:,:,:)*PETHETA(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_ThlPz, .TRUE. )
    !
!$acc end data
#endif
    CALL SECOND_MNH(ZTIME2)
    XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
  END IF
!
  IF ( KRR /= 0 ) THEN
!
!*       4.3  <THl Rnp>    
!
!
    ! Compute the turbulent variance F and F' at time t-dt.
#ifndef MNH_OPENACC
    ZF      (:,:,:) = XCTV*PLM(:,:,:)*PLEPS(:,:,:)*MZF(0.5*(PPHI3(:,:,:)+PPSI3(:,:,:))*PDTH_DZ(:,:,:)*PDR_DZ(:,:,:))
#else
!$acc kernels present_cr(ZTMP1_DEVICE)
!dir$ concurrent    
    ZTMP1_DEVICE(:,:,:) = 0.5*(PPHI3(:,:,:)+PPSI3(:,:,:))*PDTH_DZ(:,:,:)*PDR_DZ(:,:,:)
!$acc end kernels
    CALL MZF_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP2_DEVICE(:,:,:) )
!$acc kernels 
#ifdef MNH_COMPILER_NVHPC
!$acc loop independent collapse(3)    
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
      ZF      (JI,JJ,JK) = XCTV*PLM(JI,JJ,JK)*PLEPS(JI,JJ,JK)*ZTMP2_DEVICE(JI,JJ,JK)
    END DO
!$acc end kernels    
#endif
!$acc kernels present_cr(ZDFDDTDZ,ZDFDDRDZ)  
    ZDFDDTDZ(:,:,:) = 0.     ! this term, because of discretization, is treated separately
    ZDFDDRDZ(:,:,:) = 0.     ! this term, because of discretization, is treated separately
!$acc end kernels
    !
    ! Effect of 3rd order terms in temperature flux (at mass point)
    !
    ! d(w'th'2)/dz
#ifndef MNH_OPENACC
    IF (GFTH2) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_THR_WTH2(PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTH2(:,:,:)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_THR_WTH2_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTH2(:,:,:)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_THR_WTH2_O_DDRDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:)) * PFTH2(:,:,:)
    END IF
#else
    IF (GFTH2) THEN
      CALL M3_THR_WTH2( PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTH2(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WTH2_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFTH2(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WTH2_O_DDRDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFTH2(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'2th')/dz
#ifndef MNH_OPENACC
    IF (GFWTH) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_THR_W2TH(PREDR1(:,:,:),PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
       PDR_DZ(:,:,:)) * MZF(PFWTH(:,:,:))
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_THR_W2TH_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
       PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PDR_DZ(:,:,:),PETHETA(:,:,:)) * MZF(PFWTH(:,:,:))
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_THR_W2TH_O_DDRDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
       PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:)) * MZF(PFWTH(:,:,:))
    END IF
#else
    IF (GFWTH) THEN
      CALL M3_THR_W2TH( PREDR1(:,:,:),PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
       & PDR_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      CALL MZF_DEVICE( PFWTH(:,:,:), ZTMP2_DEVICE(:,:,:) )
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_W2TH_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PDR_DZ(:,:,:),PETHETA(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_W2TH_O_DDRDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
       & PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'r'2)/dz
#ifndef MNH_OPENACC
    IF (GFR2) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_THR_WR2(PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFR2(:,:,:)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_THR_WR2_O_DDTDZ(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:)) * PFR2(:,:,:)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_THR_WR2_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:)) * PFR2(:,:,:)
    END IF
#else
    IF (GFR2) THEN
      CALL M3_THR_WR2( PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
       & PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WR2_O_DDTDZ( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WR2_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
       & PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),PDTH_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
      ! d(w'2r')/dz
#ifndef MNH_OPENACC
    IF (GFWR) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_THR_W2R(PREDTH1(:,:,:),PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
      & PDTH_DZ(:,:,:)) * MZF(PFWR(:,:,:))
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_THR_W2R_O_DDTDZ(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:)) * MZF(PFWR(:,:,:))
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_THR_W2R_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PDTH_DZ(:,:,:),PEMOIST(:,:,:)) * MZF(PFWR(:,:,:))
    END IF
#else
    IF (GFWR) THEN
      CALL MZF_DEVICE( PFWR(:,:,:), ZTMP1_DEVICE(:,:,:) )
      CALL M3_THR_W2R( PREDTH1(:,:,:),PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
      & PDTH_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_W2R_O_DDTDZ( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_W2R_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PDTH_DZ(:,:,:),PEMOIST(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'th'r')/dz
#ifndef MNH_OPENACC
    IF (GFTHR) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_THR_WTHR(PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
      & PSQRT_TKE(:,:,:)) * PFTHR(:,:,:)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + D_M3_THR_WTHR_O_DDTDZ(PREDTH1(:,:,:),PREDR1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:)) * PFTHR(:,:,:)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_THR_WTHR_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:)) * PFTHR(:,:,:)
    END IF
#else
    IF (GFTHR) THEN
      CALL M3_THR_WTHR( PREDTH1(:,:,:),PREDR1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
      & PSQRT_TKE(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTHR(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WTHR_O_DDTDZ( PREDTH1(:,:,:),PREDR1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDTDZ)
      ZDFDDTDZ(:,:,:) = ZDFDDTDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFTHR(:,:,:)
      !$acc end kernels
      CALL D_M3_THR_WTHR_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFTHR(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
#ifndef MNH_OPENACC
    ZFLXZ(:,:,:)   = ZF(:,:,:)                                                     &
        + PIMPL * XCTV*PLM(:,:,:)*PLEPS(:,:,:)*0.5                                        &
          * MZF( ( D_PHI3DTDZ_O_DDTDZ(PPHI3(:,:,:),PREDTH1(:,:,:),PREDR1(:,:,:), &
                                                  PRED2TH3(:,:,:),PRED2THR3(:,:,:),HTURBDIM,GUSERV) & ! d(phi3*dthdz)/ddthdz term
                  +D_PSI3DTDZ_O_DDTDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:), &
                                      PRED2THR3(:,:,:),HTURBDIM,GUSERV) & ! d(psi3*dthdz)/ddthdz term
                 ) *PDR_DZ(:,:,:)  *DZM(PTHLP(:,:,:) - PTHLM(:,:,:)       ) / PDZZ(:,:,:)               &
                +( D_PHI3DRDZ_O_DDRDZ(PPHI3(:,:,:),PREDTH1(:,:,:),PREDR1(:,:,:),PRED2TH3(:,:,:), &
                                      PRED2THR3(:,:,:),HTURBDIM,GUSERV) & ! d(phi3*drdz )/ddrdz term
                  +D_PSI3DRDZ_O_DDRDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:), &
                                      PRED2THR3(:,:,:),HTURBDIM,GUSERV) & ! d(psi3*drdz )/ddrdz term
                 ) *PDTH_DZ(:,:,:) *DZM(PRP(:,:,:)   - PRM(:,:,:,1)) / PDZZ(:,:,:)               &
               )                                                            &
        + PIMPL * ZDFDDTDZ(:,:,:) * MZF(DZM(PTHLP(:,:,:) - PTHLM(:,:,:)) / PDZZ(:,:,:) )         &
        + PIMPL * ZDFDDRDZ(:,:,:) * MZF(DZM(PRP(:,:,:)   - PRM(:,:,:,1)) / PDZZ(:,:,:) )
#else
    !$acc kernels present_cr(ZTMP1_DEVICE,ZTMP2_DEVICE)
    ZTMP1_DEVICE(:,:,:) = PTHLP(:,:,:) - PTHLM(:,:,:)
    ZTMP2_DEVICE(:,:,:) = PRP(:,:,:)   - PRM(:,:,:,1)
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP3_DEVICE(:,:,:) )
    CALL DZM_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP4_DEVICE(:,:,:) )
    !$acc kernels 
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP1_DEVICE(JI,JJ,JK) = ZTMP3_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP4_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)      
    END DO

    !$acc end kernels
    CALL D_PHI3DTDZ_O_DDTDZ(PPHI3(:,:,:),PREDTH1(:,:,:),PREDR1(:,:,:),PRED2TH3(:,:,:),PRED2THR3(:,:,:), &
                            HTURBDIM,GUSERV,ZTMP3_DEVICE(:,:,:)) ! d(phi3*dthdz)/ddthdz term
    CALL D_PSI3DTDZ_O_DDTDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:),PRED2THR3(:,:,:), &
                            HTURBDIM,GUSERV,ZTMP4_DEVICE(:,:,:))  ! d(psi3*dthdz)/ddthdz term
    CALL D_PHI3DRDZ_O_DDRDZ(PPHI3(:,:,:),PREDTH1(:,:,:),PREDR1(:,:,:),PRED2TH3(:,:,:),PRED2THR3(:,:,:), &
                            HTURBDIM,GUSERV,ZTMP5_DEVICE(:,:,:)) ! d(phi3*drdz )/ddrdz term
    CALL D_PSI3DRDZ_O_DDRDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:),PRED2THR3(:,:,:), &
                            HTURBDIM,GUSERV,ZTMP6_DEVICE(:,:,:))  ! d(psi3*drdz )/ddrdz term
    !$acc kernels present_cr(ZTMP1_DEVICE,ZTMP8_DEVICE)
    ZTMP1_DEVICE(:,:,:) = PTHLP(:,:,:) - PTHLM(:,:,:)
    ZTMP8_DEVICE(:,:,:) = PRP(:,:,:)   - PRM(:,:,:,1)
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP2_DEVICE(:,:,:) )
    CALL DZM_DEVICE( ZTMP8_DEVICE(:,:,:), ZTMP1_DEVICE(:,:,:) )
    !!!
    !$acc kernels
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP7_DEVICE(JI,JJ,JK) =   ( ZTMP3_DEVICE(JI,JJ,JK) + ZTMP4_DEVICE(JI,JJ,JK)) * PDR_DZ(JI,JJ,JK)  &
                              * ZTMP2_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK) &
                              + ( ZTMP5_DEVICE(JI,JJ,JK) + ZTMP6_DEVICE(JI,JJ,JK)) * PDTH_DZ(JI,JJ,JK) &
                              * ZTMP1_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
    END DO
    !$acc end kernels
    !!!
    !$acc kernels present_cr(ZTMP1_DEVICE,ZTMP2_DEVICE)
    ZTMP1_DEVICE(:,:,:) = PTHLP(:,:,:) - PTHLM(:,:,:)
    ZTMP2_DEVICE(:,:,:) = PRP(:,:,:)   - PRM(:,:,:,1)
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP3_DEVICE(:,:,:) )
    CALL DZM_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP4_DEVICE(:,:,:) )
    !$acc kernels
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP1_DEVICE(JI,JJ,JK) = ZTMP3_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP4_DEVICE(JI,JJ,JK) /PDZZ(JI,JJ,JK)
    END DO
    !$acc end kernels
    !!!
    CALL MZF_DEVICE( ZTMP7_DEVICE(:,:,:), ZTMP3_DEVICE(:,:,:) )
    CALL MZF_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP4_DEVICE(:,:,:) )
    CALL MZF_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP5_DEVICE(:,:,:) )
    !$acc kernels  present_cr(zflxz)
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZFLXZ(JI,JJ,JK)   = ZF(JI,JJ,JK)                           &
            + PIMPL * XCTV*PLM(JI,JJ,JK)*PLEPS(JI,JJ,JK)*0.5 * ZTMP3_DEVICE(JI,JJ,JK)  &
            + PIMPL * ZDFDDTDZ(JI,JJ,JK) * ZTMP4_DEVICE(JI,JJ,JK)            &
            + PIMPL * ZDFDDRDZ(JI,JJ,JK) * ZTMP5_DEVICE(JI,JJ,JK)
    END DO
#endif
    !
    ! special case near the ground ( uncentred gradient )
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(2)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
    ZFLXZ(JI,JJ,IKB) =                                            & 
    (XCHT1 * PPHI3(JI,JJ,IKB+KKL) + XCHT2 * PPSI3(JI,JJ,IKB+KKL))   &
    *( PEXPL *                                                  &
       ( ZCOEFF(JI,JJ,IKB+2*KKL)*PTHLM(JI,JJ,IKB+2*KKL)             &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PTHLM(JI,JJ,IKB+KKL  )             & 
        +ZCOEFF(JI,JJ,IKB      )*PTHLM(JI,JJ,IKB      ))            &
      *( ZCOEFF(JI,JJ,IKB+2*KKL)*PRM(JI,JJ,IKB+2*KKL,1)             &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PRM(JI,JJ,IKB+KKL,1  )             & 
        +ZCOEFF(JI,JJ,IKB      )*PRM(JI,JJ,IKB  ,1    ))            &
      +PIMPL *                                                  &
       ( ZCOEFF(JI,JJ,IKB+2*KKL)*PTHLP(JI,JJ,IKB+2*KKL)             &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PTHLP(JI,JJ,IKB+KKL  )             &
        +ZCOEFF(JI,JJ,IKB      )*PTHLP(JI,JJ,IKB      ))            &
      *( ZCOEFF(JI,JJ,IKB+2*KKL)*PRP(JI,JJ,IKB+2*KKL  )             &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PRP(JI,JJ,IKB+KKL    )             & 
        +ZCOEFF(JI,JJ,IKB      )*PRP(JI,JJ,IKB        ))            &
        )
    END DO
!$acc end kernels
    !$acc kernels present_cr(ZFLXZ,PSIGS)    
    ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
      IF ( KRRL > 0 ) THEN
      PSIGS(:,:,:) = PSIGS(:,:,:) +     &
                     2. * PATHETA(:,:,:) * PAMOIST(:,:,:) * ZFLXZ(:,:,:)
    END IF
!$acc end kernels
    ! stores <THl Rnp>   
    IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
      TZFIELD%CMNHNAME   = 'THLRCONS_VCOR'
      TZFIELD%CSTDNAME   = ''
      TZFIELD%CLONGNAME  = 'THLRCONS_VCOR'
      TZFIELD%CUNITS     = 'K kg kg-1'
      TZFIELD%CDIR       = 'XY'
      TZFIELD%CCOMMENT   = 'X_Y_Z_THLRCONS_VCOR'
      TZFIELD%NGRID      = 1
      TZFIELD%NTYPE      = TYPEREAL
      TZFIELD%NDIMS      = 3
      TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLXZ)
      CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ(:,:,:))
    END IF
!
! and we store in LES configuration
!
    IF (LLES_CALL) THEN
      CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
      CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_THlRt )
      CALL LES_MEAN_SUBGRID( MZF(PWM(:,:,:))*ZFLXZ(:,:,:), X_LES_RES_W_SBG_ThlRt )
      CALL LES_MEAN_SUBGRID( -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:), X_LES_SUBGRID_DISS_ThlRt )
      CALL LES_MEAN_SUBGRID( PETHETA(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_RtThv )
      CALL LES_MEAN_SUBGRID( -XA3*PBETA(:,:,:)*PETHETA(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_RtPz, .TRUE. )
      CALL LES_MEAN_SUBGRID( PEMOIST(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_ThlThv , .TRUE. )
      CALL LES_MEAN_SUBGRID( -XA3*PBETA(:,:,:)*PEMOIST(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_ThlPz, .TRUE. )
#else
!$acc data copy(X_LES_SUBGRID_THlRt,X_LES_RES_W_SBG_ThlRt,X_LES_SUBGRID_DISS_ThlRt, &
!$acc &         X_LES_SUBGRID_RtThv,X_LES_SUBGRID_RtPz,X_LES_SUBGRID_ThlThv,X_LES_SUBGRID_ThlPz)
      !
      CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_THlRt )
      !
      CALL MZF_DEVICE( PWM(:,:,:), ZTMP1_DEVICE(:,:,:) )
      !$acc kernels present_cr(ZTMP2_DEVICE)
      ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE(:,:,:), X_LES_RES_W_SBG_ThlRt )
      !
      !$acc kernels present_cr(ZTMP1_DEVICE)
      ZTMP1_DEVICE(:,:,:) = -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_DISS_ThlRt )
      !
      !$acc kernels present_cr(ZTMP1_DEVICE)
      ZTMP1_DEVICE(:,:,:) = PETHETA(:,:,:)*ZFLXZ(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_RtThv )
      !
      !$acc kernels present_cr(ZTMP1_DEVICE)
      ZTMP1_DEVICE(:,:,:) = -XA3*PBETA(:,:,:)*PETHETA(:,:,:)*ZFLXZ(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_RtPz, .TRUE. )
      !
      !$acc kernels present_cr(ZTMP1_DEVICE)
      ZTMP1_DEVICE(:,:,:) = PEMOIST(:,:,:)*ZFLXZ(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_ThlThv , .TRUE. )
      !
      !$acc kernels present_cr(ZTMP1_DEVICE)
      ZTMP1_DEVICE(:,:,:) = -XA3*PBETA(:,:,:)*PEMOIST(:,:,:)*ZFLXZ(:,:,:)
      !$acc end kernels
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_ThlPz, .TRUE. )
      !
!$acc end data
#endif
      CALL SECOND_MNH(ZTIME2)
      XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
    END IF
! 
!
!*       4.4  <Rnp Rnp>
!
!
    ! Compute the turbulent variance F and F' at time t-dt.
#ifndef MNH_OPENACC
#ifndef MNH_BITREP
    ZF      (:,:,:) = XCTV*PLM(:,:,:)*PLEPS(:,:,:)*MZF(PPSI3(:,:,:)*PDR_DZ(:,:,:)**2)
#else
    ZF      (:,:,:) = XCTV*PLM(:,:,:)*PLEPS(:,:,:)*MZF(PPSI3(:,:,:)*BR_P2(PDR_DZ(:,:,:)))
#endif
#else
    !$acc kernels !!present_cr(ZTMP1_DEVICE)
#ifndef MNH_BITREP
    !dir$ concurrent
    ZTMP1_DEVICE(:,:,:) = PPSI3(:,:,:)*PDR_DZ(:,:,:)**2
#else
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP1_DEVICE(JI,JJ,JK) = PPSI3(JI,JJ,JK)*BR_P2(PDR_DZ(JI,JJ,JK))
    END DO
#endif
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP1_DEVICE(:,:,:), ZTMP2_DEVICE(:,:,:) )
    !$acc kernels
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZF      (JI,JJ,JK) = XCTV*PLM(JI,JJ,JK)*PLEPS(JI,JJ,JK)*ZTMP2_DEVICE(JI,JJ,JK)
    END DO
#endif
    ZDFDDRDZ(:,:,:) = 0.     ! this term, because of discretization, is treated separately
    !$acc end kernels
    !
    ! Effect of 3rd order terms in temperature flux (at mass point)
    !
    ! d(w'r'2)/dz
#ifndef MNH_OPENACC
    IF (GFR2) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_R2_WR2(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
      & PSQRT_TKE(:,:,:)) * PFR2(:,:,:)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_R2_WR2_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:)) * PFR2(:,:,:)
    END IF
#else
    IF (GFR2) THEN
      CALL M3_R2_WR2( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
      & PSQRT_TKE(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
      CALL D_M3_R2_WR2_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PEMOIST(:,:,:),ZTMP1_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP1_DEVICE(:,:,:) * PFR2(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    ! d(w'2r')/dz
#ifndef MNH_OPENACC
    IF (GFWR) THEN
      ZF(:,:,:)       = ZF(:,:,:)       + M3_R2_W2R(PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),PDR_DZ(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:)) * MZF(PFWR(:,:,:))
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_R2_W2R_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),GUSERV) * MZF(PFWR(:,:,:))
    END IF
#else
    IF (GFWR) THEN
      CALL MZF_DEVICE( PFWR(:,:,:), ZTMP1_DEVICE(:,:,:) )
      CALL M3_R2_W2R( PREDR1(:,:,:),PREDTH1(:,:,:),PD(:,:,:),PDR_DZ(:,:,:),&
      & PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),ZTMP2_DEVICE(:,:,:))
      !$acc kernels present_cr(ZF)
      ZF(:,:,:)       = ZF(:,:,:)       + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
      CALL D_M3_R2_W2R_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),&
      & PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),GUSERV,ZTMP3_DEVICE(:,:,:))
      !$acc kernels present_cr(ZDFDDRDZ)
      ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP3_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
    END IF
#endif
    !
    IF (KRR/=0) THEN
      ! d(w'r'2)/dz
#ifndef MNH_OPENACC
      IF (GFTH2) THEN
        ZF(:,:,:)       = ZF(:,:,:)       + M3_R2_WTH2(PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
        & PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTH2(:,:,:)
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_R2_WTH2_O_DDRDZ(PREDR1(:,:,:),&
        & PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTH2(:,:,:)
      END IF
#else
      IF (GFTH2) THEN
        CALL M3_R2_WTH2( PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),&
        & PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
        !$acc kernels present_cr(ZF)
        ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTH2(:,:,:)
        !$acc end kernels
        CALL D_M3_R2_WTH2_O_DDRDZ( PREDR1(:,:,:),&
        & PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
        !$acc kernels present_cr(ZDFDDRDZ)
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFTH2(:,:,:)
        !$acc end kernels
      END IF
#endif
      !
      ! d(w'2r')/dz
#ifndef MNH_OPENACC
      IF (GFWTH) THEN
        ZF(:,:,:)       = ZF(:,:,:)       + M3_R2_W2TH(PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
          PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * MZF(PFWTH(:,:,:))
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_R2_W2TH_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),&
          PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:))  &
          * MZF(PFWTH(:,:,:))
      END IF
#else
      IF (GFWTH) THEN
        CALL MZF_DEVICE( PFWTH(:,:,:), ZTMP1_DEVICE(:,:,:) )
        CALL M3_R2_W2TH( PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),&
        & PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
        !$acc kernels present_cr(ZF)
        ZF(:,:,:)       = ZF(:,:,:)       + ZTMP2_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
        !$acc end kernels
        CALL D_M3_R2_W2TH_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),&
        & PD(:,:,:),PLM(:,:,:),PLEPS(:,:,:),PTKEM(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP3_DEVICE(:,:,:))
        !$acc kernels present_cr(ZDFDDRDZ)
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP3_DEVICE(:,:,:) * ZTMP1_DEVICE(:,:,:)
        !$acc end kernels
      END IF
#endif
      !
      ! d(w'th'r')/dz
#ifndef MNH_OPENACC
      IF (GFTHR) THEN
        ZF(:,:,:)       = ZF(:,:,:)       + M3_R2_WTHR(PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
        & PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTHR(:,:,:)
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + D_M3_R2_WTHR_O_DDRDZ(PREDR1(:,:,:),PREDTH1(:,:,:),&
        & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:)) * PFTHR(:,:,:)
      END IF
#else
      IF (GFTHR) THEN
        CALL M3_R2_WTHR( PREDTH1(:,:,:),PD(:,:,:),PLEPS(:,:,:),&
        & PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP1_DEVICE(:,:,:))
        !$acc kernels present_cr(ZF)
        ZF(:,:,:)       = ZF(:,:,:)       + ZTMP1_DEVICE(:,:,:) * PFTHR(:,:,:)
        !$acc end kernels
        CALL D_M3_R2_WTHR_O_DDRDZ( PREDR1(:,:,:),PREDTH1(:,:,:),&
        & PD(:,:,:),PLEPS(:,:,:),PSQRT_TKE(:,:,:),PBLL_O_E(:,:,:),PETHETA(:,:,:),PDR_DZ(:,:,:),ZTMP2_DEVICE(:,:,:))
        !$acc kernels present_cr(ZDFDDRDZ)
        ZDFDDRDZ(:,:,:) = ZDFDDRDZ(:,:,:) + ZTMP2_DEVICE(:,:,:) * PFTHR(:,:,:)
        !$acc end kernels
      END IF
#endif
  
    END IF
    !
#ifndef MNH_OPENACC
    ZFLXZ(:,:,:)   = ZF(:,:,:)                                                              &
          + PIMPL * XCTV*PLM(:,:,:)*PLEPS(:,:,:)                                                   &
            *MZF(D_PSI3DRDZ2_O_DDRDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:),PRED2THR3(:,:,:), &
                                                 PDR_DZ(:,:,:),HTURBDIM,GUSERV)    &
                 *DZM(PRP(:,:,:) - PRM(:,:,:,1)) / PDZZ(:,:,:)                                   ) &
          + PIMPL * ZDFDDRDZ(:,:,:) * MZF(DZM(PRP(:,:,:) - PRM(:,:,:,1)) / PDZZ(:,:,:) )
#else
    CALL D_PSI3DRDZ2_O_DDRDZ(PPSI3(:,:,:),PREDR1(:,:,:),PREDTH1(:,:,:),PRED2R3(:,:,:),PRED2THR3(:,:,:),PDR_DZ(:,:,:), &
                             HTURBDIM,GUSERV,ZTMP1_DEVICE(:,:,:))
    !$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,:) = PRP(:,:,:) - PRM(:,:,:,1)
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP3_DEVICE(:,:,:) )
    !$acc kernels
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP1_DEVICE(:,:,:) )
    !$acc kernels
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)    
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP3_DEVICE(JI,JJ,JK) / PDZZ(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE(:,:,:), ZTMP3_DEVICE(:,:,:) )
#ifdef MNH_COMPILER_CCE
    !$acc kernels 
#else
    !$acc kernels 
#endif
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(3)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZFLXZ(JI,JJ,JK)   = ZF(JI,JJ,JK)                        &
            + PIMPL * XCTV*PLM(JI,JJ,JK)*PLEPS(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK) &
            + PIMPL * ZDFDDRDZ(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK)
    END DO
   !$acc end kernels
#endif
    !
    ! special case near the ground ( uncentred gradient )
#ifndef MNH_BITREP
    !$acc kernels  present_cr(ZFLXZ)
    ZFLXZ(:,:,IKB) = XCHV * PPSI3(:,:,IKB+KKL) * PLM(:,:,IKB)   &
        * PLEPS(:,:,IKB)                                        &
    *( PEXPL *                                                  &
       ( ZCOEFF(:,:,IKB+2*KKL)*PRM(:,:,IKB+2*KKL,1)             &
        +ZCOEFF(:,:,IKB+KKL  )*PRM(:,:,IKB+KKL,1  )             & 
        +ZCOEFF(:,:,IKB      )*PRM(:,:,IKB  ,1    ))**2         &
      +PIMPL *                                                  &
       ( ZCOEFF(:,:,IKB+2*KKL)*PRP(:,:,IKB+2*KKL)               &
        +ZCOEFF(:,:,IKB+KKL  )*PRP(:,:,IKB+KKL  )               &
        +ZCOEFF(:,:,IKB      )*PRP(:,:,IKB      ))**2           &
     ) 
   !$acc end kernels
#else
    !$acc kernels 
#ifdef MNH_COMPILER_NVHPC
    !$acc loop independent collapse(2)
#endif
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
    ZFLXZ(JI,JJ,IKB) = XCHV * PPSI3(JI,JJ,IKB+KKL) * PLM(JI,JJ,IKB)   &
        * PLEPS(JI,JJ,IKB)                                        &
    *( PEXPL *                                                  &
       BR_P2( ZCOEFF(JI,JJ,IKB+2*KKL)*PRM(JI,JJ,IKB+2*KKL,1)        &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PRM(JI,JJ,IKB+KKL,1  )             & 
        +ZCOEFF(JI,JJ,IKB      )*PRM(JI,JJ,IKB  ,1    ))            &
      +PIMPL *                                                  &
       BR_P2( ZCOEFF(JI,JJ,IKB+2*KKL)*PRP(JI,JJ,IKB+2*KKL)          &
        +ZCOEFF(JI,JJ,IKB+KKL  )*PRP(JI,JJ,IKB+KKL  )               &
        +ZCOEFF(JI,JJ,IKB      )*PRP(JI,JJ,IKB      ))              &
        )
    END DO
   !$acc end kernels
#endif
    !
    !$acc kernels  present_cr(ZFLXZ)
    ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
    !
    IF ( KRRL > 0 ) THEN
#ifndef MNH_BITREP
      PSIGS(:,:,:) = PSIGS(:,:,:) + PAMOIST(:,:,:) **2 * ZFLXZ(:,:,:)
#else
      PSIGS(:,:,:) = PSIGS(:,:,:) + BR_P2(PAMOIST(:,:,:)) * ZFLXZ(:,:,:)
#endif
    END IF
!$acc end kernels
    ! stores <Rnp Rnp>    
    IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
      TZFIELD%CMNHNAME   = 'RTOT_VVAR'
      TZFIELD%CSTDNAME   = ''
      TZFIELD%CLONGNAME  = 'RTOT_VVAR'
      TZFIELD%CUNITS     = 'kg2 kg-2'
      TZFIELD%CDIR       = 'XY'
      TZFIELD%CCOMMENT   = 'X_Y_Z_RTOT_VVAR'
      TZFIELD%NGRID      = 1
      TZFIELD%NTYPE      = TYPEREAL
      TZFIELD%NDIMS      = 3
      TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLXZ)
      CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ(:,:,:))
    END IF
    !
    ! and we store in LES configuration
    !
    IF (LLES_CALL) THEN
      CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
      CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_Rt2 )
      CALL LES_MEAN_SUBGRID( MZF(PWM(:,:,:))*ZFLXZ(:,:,:), X_LES_RES_W_SBG_Rt2 )
      CALL LES_MEAN_SUBGRID( PEMOIST(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_RtThv , .TRUE. )
      CALL LES_MEAN_SUBGRID( -XA3*PBETA(:,:,:)*PEMOIST(:,:,:)*ZFLXZ(:,:,:), X_LES_SUBGRID_RtPz, .TRUE. )
      CALL LES_MEAN_SUBGRID( -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:), X_LES_SUBGRID_DISS_Rt2 )
#else
!$acc data copy(X_LES_SUBGRID_Rt2,X_LES_RES_W_SBG_Rt2,X_LES_SUBGRID_RtThv, &
!$acc &         X_LES_SUBGRID_RtPz,X_LES_SUBGRID_DISS_Rt2)
      !
      CALL LES_MEAN_SUBGRID( ZFLXZ(:,:,:), X_LES_SUBGRID_Rt2 )
      !
      CALL MZF_DEVICE( PWM(:,:,:), ZTMP1_DEVICE(:,:,:) )
      ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:) * ZFLXZ(:,:,:)
      CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE(:,:,:), X_LES_RES_W_SBG_Rt2 )
      !
      ZTMP1_DEVICE(:,:,:) = PEMOIST(:,:,:)*ZFLXZ(:,:,:)
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_RtThv , .TRUE. )
      !
      ZTMP1_DEVICE(:,:,:) = -XA3*PBETA(:,:,:)*PEMOIST(:,:,:)*ZFLXZ(:,:,:)
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_RtPz, .TRUE. )
      !
      ZTMP1_DEVICE(:,:,:) = -2.*XCTD*PSQRT_TKE(:,:,:)*ZFLXZ(:,:,:)/PLEPS(:,:,:)
      CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE(:,:,:), X_LES_SUBGRID_DISS_Rt2 )
      !
!$acc end data
#endif
      CALL SECOND_MNH(ZTIME2)
      XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
    END IF
    !
  END IF  ! end if KRR ne 0
!
!
!        4.5  Vertical part of Sigma_s
!
  IF ( KRRL > 0 ) THEN
    ! Extrapolate PSIGS at the ground and at the top
!$acc kernels present_cr(PSIGS)
    PSIGS(:,:,KKA) = PSIGS(:,:,IKB)
    PSIGS(:,:,KKU) = PSIGS(:,:,IKE)
    PSIGS(:,:,:) =  SQRT( MAX (PSIGS(:,:,:) , 1.E-12) )
!$acc end kernels
  END IF

if ( mppdb_initialized ) then
  !Check all out arrays
  call Mppdb_check( psigs, "Turb_ver_thermo_corr end:psigs" )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate( zflxz, zkeff, zf , zdfddtdz , zdfddrdz , zcoeff )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data

!----------------------------------------------------------------------------
END SUBROUTINE TURB_VER_THERMO_CORR
